# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Utilities for working with dynamic sector specifications."""

import logging

from pyresample import load_area

from geoips.sector_utils.tc_tracks import set_tc_area_def
from geoips.interfaces import sectors
from geoips.errors import PluginError

LOG = logging.getLogger(__name__)

SECTOR_INFO_ATTRS = {
    "tc": [
        "pressure",
        "vmax",
        "clat",
        "clon",
        "synoptic_time",
        "aid_type",
        "storm_num",
        "storm_name",
        "storm_basin",
        "storm_year",
        "deck_line",
        "source_file",
        "final_storm_name",
    ],
    "pyrocb": ["min_lat", "min_lon", "max_lat", "max_lon", "box_resolution_km"],
    "volcano": [
        "summit_elevation",
        "plume_height",
        "wind_speed",
        "wind_dir",
        "clat",
        "clon",
    ],
    "atmosriver": ["min_lat", "min_lon", "max_lat", "max_lon", "box_resolution_km"],
    "stitched": [
        "continent",
        "country",
        "area",
        "subarea",
        "state",
        "city",
        "primary_area_definition",
    ],
    "static": ["continent", "country", "area", "subarea", "state", "city"],
}


def set_text_area_def(xarray_obj, area_def):
    """Set the area definition for text files.

    This uses raw sectored data, not interpolated.

    Parameters
    ----------
    xarray_obj : xarray.Dataset
        xarray dataset
    area_def : pyresample.AreaDefinition
        original area definition

    Returns
    -------
    pyresample.AreaDefinition
        pyresample AreaDefinition pertaining to the region for generating text file
    """
    text_area_def = area_def
    num_lines = xarray_obj.wind_speed_kts.where(
        xarray_obj.wind_speed_kts > 0, drop=True
    ).shape[0]
    num_samples = xarray_obj.wind_speed_kts.where(
        xarray_obj.wind_speed_kts > 0, drop=True
    ).shape[1]
    # Uses default pixel width/height
    text_area_def = set_tc_area_def(
        area_def.sector_info,
        num_lines=num_lines,
        num_samples=num_samples,
        pixel_width=None,
        pixel_height=None,
    )
    text_area_def.pixel_size_x = "native"
    text_area_def.pixel_size_y = "native"

    return text_area_def


def check_center_coverage(
    xarray_obj,
    area_def,
    varlist,
    covg_varname=None,
    covg_varlist=None,
    width_degrees=8,
    height_degrees=8,
    verbose=False,
    hours_before_sector_time=18,
    hours_after_sector_time=6,
):
    """Check if there is any data covering the center of the sector.

    Do not provide any longitude padding for coverage check sectoring -
    we want to see if there is any data within the exact center box,
    not within +- 3 degrees of the center box.
    """
    covg_area_def = set_tc_coverage_check_area_def(
        area_def, width_degrees=width_degrees, height_degrees=height_degrees
    )

    from geoips.xarray_utils.data import sector_xarray_dataset

    covg_xarray = sector_xarray_dataset(
        xarray_obj,
        covg_area_def,
        varlist,
        lon_pad=0,
        verbose=verbose,
        hours_before_sector_time=hours_before_sector_time,
        hours_after_sector_time=hours_after_sector_time,
    )

    from geoips.data_manipulations.info import percent_unmasked

    # If no covg_xarray returned, return False
    if covg_xarray is None:
        return False, covg_xarray

    # If we passed in a list of coverage variables, loop through each one - if
    # any are > 0, return True
    if covg_varlist is not None:
        for curr_covg_varname in covg_varlist:
            if percent_unmasked(covg_xarray[curr_covg_varname].to_masked_array()) > 0:
                return True, covg_xarray

    # If we only want a single coverage variable, return False if that variable is 0
    if covg_varname is not None:
        if percent_unmasked(covg_xarray[covg_varname].to_masked_array()) == 0:
            return False, covg_xarray

    # Otherwise, return True
    return True, covg_xarray


def copy_sector_info(src_area_def, dest_area_def):
    """Copy sector info from src_area_def to dest_area_def."""
    if hasattr(src_area_def, "sector_info"):
        dest_area_def.sector_info = src_area_def.sector_info
    if hasattr(src_area_def, "sector_type"):
        dest_area_def.sector_type = src_area_def.sector_type
    if hasattr(src_area_def, "sector_start_datetime"):
        dest_area_def.sector_start_datetime = src_area_def.sector_start_datetime
    if hasattr(src_area_def, "sector_end_datetime"):
        dest_area_def.sector_end_datetime = src_area_def.sector_end_datetime
    return dest_area_def


def set_tc_coverage_check_area_def(area_def, width_degrees=8, height_degrees=8):
    """Set the area definition for checking coverage for TC overpasses.

    Take a small box around the center of the storm to evaluate coverage,
    rather than the entire image.

    Parameters
    ----------
    area_def : pyresample.AreaDefinition
        original area definition

    Returns
    -------
    pyresample.AreaDefinition
        pyresample AreaDefinition pertaining to the region for plotting
    """
    covg_area_def = area_def
    DEG_TO_KM = 111.321
    # Take a 8deg x 8deg box centered over the storm location for
    # coverage check, 1km pixels
    width_km = DEG_TO_KM * width_degrees
    height_km = DEG_TO_KM * height_degrees
    LOG.info("  Changing area definition for checking TC coverage")

    from geoips.plugins.modules.sector_spec_generators import center_coordinates

    covg_area_def = center_coordinates.call(
        area_id=area_def.area_id,
        long_description=area_def.description,
        clat=area_def.sector_info["clat"],
        clon=area_def.sector_info["clon"],
        projection="eqc",
        num_lines=int(height_km),
        num_samples=int(width_km),
        pixel_width=1000.0,
        pixel_height=1000.0,
    )
    covg_area_def = copy_sector_info(area_def, covg_area_def)

    LOG.info("  Coverage area definition: %s", covg_area_def.name)
    LOG.info(
        "  Coverage sector info: clat: %s clon: %s",
        covg_area_def.sector_info["clat"],
        covg_area_def.sector_info["clon"],
    )

    return covg_area_def


def get_max_lat(lats):
    """Get maximum latitude from array of latitudes.

    Parameters
    ----------
    lats : numpy.ndarray
        numpy MaskedArray of latitudes

    Returns
    -------
    float
        Maximum latitude, between -90 and 90
    """
    return float(lats.max())


def get_min_lat(lats):
    """Get minimum latitude from array of latitudes.

    Parameters
    ----------
    lats : numpy.ndarray
        numpy MaskedArray of latitudes

    Returns
    -------
    float
        Minimum latitude, between -90 and 90
    """
    return float(lats.min())


def get_min_lon(lons):
    """Get minimum longitude from array of longitudes, handling date line.

    Parameters
    ----------
    lons : numpy.ndarray
        numpy MaskedArray of longitudes

    Returns
    -------
    float
        Minimum longitude, between -180 and 180
    """
    if lons.max() > 179.5 and lons.min() < -179.5:
        import numpy

        lons = numpy.ma.where(lons < 0, lons + 360, lons)
    minlon = lons.min()
    if minlon > 180:
        minlon -= 360
    return float(minlon)


def get_max_lon(lons):
    """Get maximum longitude from array of longitudes, handling date line.

    Parameters
    ----------
    lons : numpy.ndarray
        numpy MaskedArray of longitudes

    Returns
    -------
    float
        Maximum longitude, between -180 and 180
    """
    if lons.max() > 179.5 and lons.max() < -179.5:
        import numpy

        lons = numpy.ma.where(lons < 0, lons + 360, lons)
    maxlon = lons.max()
    if maxlon > 180:
        maxlon -= 360
    return float(maxlon)


def get_lat_center(lats):
    """Return the center longitude point from lats array."""
    center_lat = lats.min() + (lats.max() - lats.min()) / 2.0
    return center_lat


def get_lon_center(lons):
    """Return the center longitude point from lons array."""
    import numpy

    if lons.max() > 179.5 and lons.min() < -179.5:
        lons = numpy.ma.where(lons < 0, lons + 360, lons)

    center_lon = lons.min() + (lons.max() - lons.min()) / 2.0

    if center_lon > 180:
        center_lon -= 360

    return center_lon


def get_trackfile_area_defs(
    trackfiles,
    trackfile_parser,
    trackfile_sectorlist=None,
    tc_spec_template=None,
    aid_type=None,
    start_datetime=None,
    end_datetime=None,
):
    """Get all TC area definitions for the current xarray object, and requested sectors.

    Parameters
    ----------
    trackfiles : list
        List of trackfiles to convert into area_defs
    trackfile_parser : str
        Parser to use from plugins.modules.sector_metadata_generators on trackfiles
    trackfile_sectorlist list of str, default=None
        * list of sector names to process, of format: tc2020io01amphan.
        * If None, or 'all' contained in list, process all matching TC sectors.
    aid_type : str, default=None
        If specified, string to look for in "aid_type" TC deck file field for
        inclusion

    Returns
    -------
    list of pyresample.AreaDefinition
        List of pyresample AreaDefinition objects
    """
    area_defs = []
    final_area_defs = []
    from geoips.sector_utils.tc_tracks import trackfile_to_area_defs

    for trackfile in trackfiles:
        area_defs += trackfile_to_area_defs(
            trackfile,
            trackfile_parser=trackfile_parser,
            tc_spec_template=tc_spec_template,
        )
    if trackfile_sectorlist is not None and "all" not in trackfile_sectorlist:
        for area_def in area_defs:
            if area_def.area_id in trackfile_sectorlist:
                final_area_defs += [area_def]
            else:
                LOG.info(
                    "area_def %s not in trackfile_sectorlist %s, not including",
                    area_def.area_id,
                    str(trackfile_sectorlist),
                )
    else:
        final_area_defs = area_defs

    ret_area_defs = final_area_defs
    if start_datetime is not None and end_datetime is not None:
        ret_area_defs = []
        for area_def in final_area_defs:
            if (
                area_def.sector_start_datetime > start_datetime
                and area_def.sector_start_datetime < end_datetime
            ):
                ret_area_defs += [area_def]

    # Make sure there are no duplicates
    ret_area_defs = remove_duplicate_storm_positions(ret_area_defs, aid_type)

    return ret_area_defs


def get_static_area_defs_for_xarray(xarray_obj, sectorlist):
    """Get all STATIC area definitions for the current xarray object.

    Filter based on requested sectors.

    Parameters
    ----------
    xarray_obj : xarray.Dataset
        xarray Dataset to which we are assigning area_defs
    sectorlist : list of str
        list of sector names

    Returns
    -------
    list of pyresample.AreaDefinition
        List of pyresample AreaDefinition objects
    """
    area_defs = []
    if (
        xarray_obj is not None
        and "area_definition" in xarray_obj.attrs.keys()
        and xarray_obj.area_definition is not None
    ):
        if xarray_obj.area_definition.area_id in sectorlist:
            LOG.info(
                "%s area_id in sectorlist and in xarray_obj.area_definition, adding "
                "area_def",
                xarray_obj.area_definition.area_id,
            )
            area_defs = [xarray_obj.area_definition]
        else:
            LOG.info(
                "%s area_id NOT in sectorlist and set on xarray_obj.area_definition, "
                "SKIPPING area_def",
                xarray_obj.area_definition.area_id,
            )
    elif sectorlist is not None:
        area_defs = get_sectors_from_yamls(sectorlist)

    ret_area_defs = []
    for area_def in area_defs:
        if area_def.name not in [curr_area_def.name for curr_area_def in ret_area_defs]:
            LOG.info("Including area_def %s in return list", area_def.name)
            ret_area_defs += [area_def]
        else:
            LOG.info("area_def %s already in return list, not including", area_def.name)

    return ret_area_defs


def get_tc_area_defs_for_xarray(
    xarray_obj,
    tcdb_sector_list=None,
    tc_spec_template=None,
    trackfile_parser=None,
    hours_before_sector_time=18,
    hours_after_sector_time=6,
    aid_type=None,
):
    """Get all TC area definitions for the current xarray object, and requested sectors.

    Parameters
    ----------
    xarray_obj : xarray.Dataset
        xarray Dataset to which we are assigning area_defs
    tcdb_sector_list : list of str, default=None
        * list of sector names to process, of format: tc2020io01amphan.
        * If None, or 'all' contained in list, process all matching TC sectors.
    actual_datetime : datetime.datetime, default=None
        Optional datetime to match for dynamic sectors
    var_for_coverage : str
        Default None, optional variable to sector to check exact time
    hours_before_sector_time : float, default=18
        hours to look before sector time
    hours_after_sector_time : float, default=6
        hours to look after sector time
    aid_type : str, default=None
        string to look for in "aid_type" TC deck file field for inclusion

    Returns
    -------
    list of pyresample AreaDefinition
        List of pyresample AreaDefinition objects required for passed xarray
    """
    if (
        "area_definition" in xarray_obj.attrs.keys()
        and xarray_obj.area_definition is not None
    ):
        return [xarray_obj.area_definition]

    area_defs = []
    from datetime import timedelta
    from geoips.sector_utils.tc_tracks_database import get_all_storms_from_db

    curr_area_defs = get_all_storms_from_db(
        xarray_obj.start_datetime - timedelta(hours=hours_before_sector_time),
        xarray_obj.end_datetime + timedelta(hours=hours_after_sector_time),
        tc_spec_template=tc_spec_template,
        trackfile_parser=trackfile_parser,
    )
    if tcdb_sector_list is not None and "all" not in tcdb_sector_list:
        for area_def in curr_area_defs:
            if area_def.area_id in tcdb_sector_list:
                area_defs += [area_def]
            else:
                LOG.info(
                    "area_def %s not in tcdb_sector_list %s, not including",
                    area_def.area_id,
                    str(tcdb_sector_list),
                )
    else:
        area_defs = curr_area_defs

    # Make sure there are no duplicates
    ret_area_defs = remove_duplicate_storm_positions(area_defs, aid_type)
    return ret_area_defs


def is_requested_aid_type(area_def, aid_type=None):
    """Return True if passed area_def is of requested aid_type."""
    if (
        aid_type is not None
        and "aid_type" in area_def.sector_info
        and area_def.sector_info["aid_type"] != aid_type
    ):
        return False
    return True


def storm_locations_match(area_def, other_area_def):
    """Return True if passed pyresample AreaDefinitions are the same location.

    Match if center lat, center lon, storm year, storm basin, and synoptic time
    all match.
    """
    if (
        area_def.sector_info["clat"] == other_area_def.sector_info["clat"]
        and area_def.sector_info["clon"] == other_area_def.sector_info["clon"]
        and area_def.sector_info["storm_year"]
        == other_area_def.sector_info["storm_year"]
        and area_def.sector_info["storm_basin"]
        == other_area_def.sector_info["storm_basin"]
        and area_def.sector_info["synoptic_time"]
        == other_area_def.sector_info["synoptic_time"]
    ):
        return True
    return False


def remove_duplicate_storm_positions(area_defs, aid_type=None):
    """Remove duplicate storm positions from passed list of area_defs.

    Uses "is_requested_aid_type" and "storm_locations_match" utilities.
    """
    ret_area_defs = []
    for area_def in area_defs:
        if not is_requested_aid_type(area_def, aid_type):
            LOG.debug(
                "area_def %s aid_type %s not requested, not including",
                area_def.name,
                area_def.sector_info["aid_type"],
            )
            continue
        elif area_def.name in [curr_area_def.name for curr_area_def in ret_area_defs]:
            LOG.info("area_def %s already in return list, not including", area_def.name)
            continue
        else:
            kept_one = False
            for other_area_def in area_defs:
                if (
                    area_def.sector_info["final_storm_name"].lower() == "invest"
                    and other_area_def.sector_info["final_storm_name"].lower()
                    != "invest"
                    and storm_locations_match(area_def, other_area_def)
                    and is_requested_aid_type(other_area_def, aid_type)
                ):
                    kept_one = True
                    LOG.info(
                        "Including area_def %s in return list, NOT including %s",
                        other_area_def.name,
                        area_def.name,
                    )
                    ret_area_defs += [other_area_def]
                if (
                    "invest_storm_id" in other_area_def.sector_info
                    and other_area_def.sector_info["invest_storm_id"]
                    == area_def.area_id
                ):
                    kept_one = True
                    LOG.info(
                        "Including area_def %s in return list, NOT including %s, "
                        "invest_storm_id defined",
                        other_area_def.name,
                        area_def.name,
                    )
                    ret_area_defs += [other_area_def]

            if not kept_one and "aid_type" in area_def.sector_info:
                ret_area_defs += [area_def]
                LOG.debug(
                    "Including area_def %s in return list, track type %s",
                    area_def.name,
                    area_def.sector_info["aid_type"],
                )
            elif not kept_one:
                ret_area_defs += [area_def]
                LOG.info("Including area_def %s in return list", area_def.name)

    return ret_area_defs


def filter_area_defs_actual_time(area_defs, actual_datetime):
    """Filter list of area_defs to only include the passed actual_datetime."""
    ret_area_def_ids = {}
    for area_def in area_defs:
        if area_def.area_id not in ret_area_def_ids:
            ret_area_def_ids[area_def.area_id] = area_def
        elif is_dynamic_sector(area_def) and actual_datetime is not None:
            if abs(actual_datetime - area_def.sector_start_datetime) < abs(
                actual_datetime
                - ret_area_def_ids[area_def.area_id].sector_start_datetime
            ):
                LOG.debug(
                    "AREA_DEF LIST REPLACING %s with area_def %s",
                    ret_area_def_ids[area_def.area_id].name,
                    area_def.name,
                )
                ret_area_def_ids[area_def.area_id] = area_def
        else:
            LOG.warning(
                "AREA_DEF LIST REPLACING Multiple identical sectors - using latest %s",
                area_def.name,
            )
            ret_area_def_ids[area_def.area_id] = area_def

    return ret_area_def_ids.values()


# def filter_area_defs_sector(area_defs, xarray_obj, var_for_coverage):
#     from geoips.xarray_utils.data import sector_xarrays
#     ret_area_def_ids = {}
#     ret_area_def_sects = {}
#     for area_def in area_defs:
#         if area_def.area_id not in ret_area_def_ids:
#             sects = sector_xarrays([xarray_obj],
#                                    area_def,
#                                    varlist=[var_for_coverage],
#                                    verbose=False)
#             if not sects:
#                 LOG.info('AREA_DEF LIST NO COVERAGE not adding sector')
#             else:
#                 LOG.info('AREA_DEF LIST ADDING area_def %s', area_def.name)
#                 ret_area_def_ids[area_def.area_id] = area_def
#                 ret_area_def_sects[area_def.area_id] = sects[0]
#         elif is_dynamic_sector(area_def):
#             old_sect = ret_area_def_sects[area_def.area_id]
#             old_area_def = ret_area_def_ids[area_def.area_id]
#             sects = sector_xarrays([xarray_obj],
#                                    area_def,
#                                    varlist=[var_for_coverage],
#                                    verbose=False)
#             if sects and abs(sects[0].start_datetime - area_def.sector_start_datetime)
#                   < abs(old_sect.start_datetime - old_area_def.sector_start_datetime):
#
#                 LOG.info('AREA_DEF LIST REPLACING %s with area_def %s',
#                          ret_area_def_ids[area_def.area_id].name,
#                          area_def.name)
#                 ret_area_def_ids[area_def.area_id] = area_def
#                 ret_area_def_sects[area_def.area_id] = sects[0]
#
#         else:
#             LOG.warning('AREA_DEF LIST REPLACING Multiple identical sectors - using
#                         latest %s', area_def.name)
#             ret_area_def_ids[area_def.area_id] = area_def
#
#     return ret_area_def_ids.values()


def get_sectors_from_yamls(sector_list):
    """Get AreaDefinition objects with custom "sector_info" dictionary.

    Based on YAML area definition contained in "sectorfnames" files.

    Parameters
    ----------
    sector_list : list of str
        list of strings of desired sector names to retrieve from YAML files

    Returns
    -------
    list
        List of pyresample AreaDefinition objects, with arbitrary additional YAML
        entries added as attributes to each area def (this is to allow specifying
        "sector_info" metadata dictionary within the YAML file)
    """
    from importlib.resources import files

    area_defs = []
    for sector_name in sector_list:
        try:
            sector_plugin = sectors.get_plugin(sector_name)
        except PluginError as resp:
            raise PluginError(
                f"{resp}:\nError getting sector {sector_name} from "
                f"\nsector_list {sector_list}. "
                f"\nCheck plugin directories for sector plugin named "
                f"{sector_name}"
            )
        abspath = str(files(sector_plugin.package) / sector_plugin.relpath)
        area_def = load_area(abspath, "spec")
        area_def.__setattr__("sector_info", sector_plugin["metadata"])
        area_def.__setattr__("sector_type", sector_plugin["family"])
        area_defs += [area_def]
    return area_defs


def create_areadefinition_from_yaml(yamlfile, sector):
    """Take a YAML with misc metadata and create a pyresample areadefinition.

    Misc. metadata will be parsed from the YAML file and manually added to the
    areadefinition

    Parameters
    ----------
    yamlfile : str
        full path to YAML area definition file
    sector : str
        name of sector

    Returns
    -------
    pyresample.AreaDefinition
        pyresample AreaDefinition based on YAML specification.
    """
    import pyresample
    import yaml

    with open(yamlfile, "r") as f:
        sectorfile_yaml = yaml.safe_load(f)
    sector_info = sectorfile_yaml["spec"]
    area_id = sector
    description = sector_info.pop("description")
    projection = sector_info.pop("projection")
    resolution = sector_info.pop("resolution")
    shape = sector_info.pop("shape")
    area_extent = sector_info.pop("area_extent")
    # Create the pyresample area definition
    shape_list = [shape["height"], shape["width"]]
    extent_list = []
    extent_list.extend(area_extent["lower_left_xy"])
    extent_list.extend(area_extent["upper_right_xy"])
    area_def = pyresample.create_area_def(
        area_id=area_id,
        description=description,
        projection=projection,
        resolution=resolution,
        shape=shape_list,
        area_extent=extent_list,
    )
    # Manually add the metadata to area_def
    for key in sector_info.keys():
        area_def.__setattr__(key, sector_info[key])
    return area_def


def is_sector_type(area_def, sector_type_str):
    """Determine if the type of area_def sector is as specified in passed sector_type.

    Parameters
    ----------
    area_def : pyresample.AreaDefinition
        pyresample AreaDefinition object specifying region of interest
    sector_type_str : str
        * String specifying the type of sector, must match 'sector_type'
          attribute on AreaDefinition object
        * currently one of 'tc', 'pyrocb', 'volcano', 'atmosriver' 'static'

    Returns
    -------
    bool
        True if area_def.sector_type == 'sector_type', False otherwise
    """
    if not area_def:
        LOG.info("area_def not defined, not of type %s", sector_type_str)
        return False
    if not hasattr(area_def, "sector_type"):
        LOG.info("area_def.sector_type not defined, not of type %s", sector_type_str)
        return False

    if area_def.sector_type == sector_type_str:
        # Need to decide if this is necessary.
        # for attr in SECTOR_INFO_ATTRS[sector_type_str]:
        #     if attr not in area_def.sector_info.keys():
        #         LOG.warning('attr %s not in area_def.sector_info.keys(), not of type
        #                     %s', attr, sector_type_str)
        #         return False
        return True
    return False


def is_dynamic_sector(area_def):
    """Determine if the AreaDefinition object is a dynamic region of interest.

    Parameters
    ----------
    area_def : pyresample.AreaDefinition
        pyresample AreaDefinition object specifying region of interest

    Returns
    -------
    bool
        * True if area_def.sector_start_datetime exists and is not None,
        * False otherwise
    """
    if not area_def:
        return False
    if not hasattr(area_def, "sector_start_datetime"):
        return False
    if not area_def.sector_start_datetime:
        return False
    if area_def.sector_start_datetime is not None:
        return True
    return False



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Apply min/max values, normalize, and invert data arrays."""
# Python Standard Libraries
import logging

# Installed Libraries
import numpy
from xarray import DataArray

LOG = logging.getLogger(__name__)


def mask_day(data_array, sunzen_array, max_zenith=90):
    """Mask where solar zenith angle less than the maxinum specified value.

    Mask all pixels within the data array where the solar zenith angle is less
    than the maxinum specified value.

    Parameters
    ----------
    data_array : numpy.ndarray or numpy.ma.MaskedArray
        data values to be masked
    sunzen_array : numpy.ndarray
        numpy.ndarray or numpy.ma.MaskedArray of solar zenith angles,
        of the same shape as the data array
    max_zenith : float, optional
        Mask all locations in data_array where sunzen_array is less
        than max_zenith, by default 90

    Returns
    -------
    numpy.ma.MaskedArray
        Data array with all locations corresponding to a
        solar zenith angle less than max_zenith masked.
    """
    LOG.info(
        "CORRECTION Masking day less than sun zenith angle of %s, "
        "min sun zen in array %s, max sun zen %s",
        max_zenith,
        sunzen_array.min(),
        sunzen_array.max(),
    )
    return numpy.ma.masked_where(sunzen_array < max_zenith, data_array)


def mask_night(data_array, sunzen_array, min_zenith=90):
    """Mask where solar zenith angle greater than the minimum specified value.

    Mask all pixels within the data array where the solar zenith angle is
    greater than the mininum specified value.

    Parameters
    ----------
    data_array : numpy.ndarray or numpy.ma.MaskedArray
        data values to be masked.
    sunzen_array : numpy.ndarray or numpy.ma.MaskedArray
        array of solar zenith angles, same shape as the data array.
    min_zenith : float, optional
        Mask all locations in data_array where sunzen_array is greater than
        min_zenith, by default 90.

    Returns
    -------
    numpy.ma.MaskedArray
        Data array with all locations corresponding to a
        solar zenith angle greater than min_zenith masked.
    """
    LOG.info(
        "CORRECTION Masking night greater than sun zenith angle of %s, "
        "min sun zen in array %s, max sun zen %s",
        min_zenith,
        sunzen_array.min(),
        sunzen_array.max(),
    )
    return numpy.ma.masked_where(sunzen_array > min_zenith, data_array)


def apply_gamma(data_array, gamma):
    """Apply gamma correction to all values in the data array.

    Gamma correction applied as: data_array ** (1.0 / float(gamma))

    Parameters
    ----------
    data_array : numpy.ndarray or numpy.ma.MaskedArray
        data array to which gamma will be applied
    gamma : float
        gamma correction value

    Returns
    -------
    numpy.ndarray
        Return numpy.ndarray or numpy.ma.MaskedArray
        if data_array was MaskedArray with gamma correction applied
        data_array ** (1.0 / float(gamma))
    """
    LOG.info("CORRECTION Applying gamma correction of %s", gamma)
    return data_array ** (1.0 / float(gamma))


def apply_solar_zenith_correction(data_array, sunzen_array):
    """Apply solar zenith angle correction to all values in data_array.

    Solar zenith correction applied as: data / cos(sunzen)

    Parameters
    ----------
    data_array : numpy.ndarray or numpy.ma.MaskedArray
        data values to be masked
    sunzen_array : numpy.ndarray or numpy.ma.MaskedArray
        solar zenith angles of the same shape as the data array.

    Returns
    -------
    numpy.ndarray
        Return numpy.ndarray or numpy.ma.MaskedArray if original
        data_array was MaskedArray with each value in the
        data_array divided by cos(sunzen).
    """
    origmask = data_array.mask
    LOG.info(
        "CORRECTION Applying solar zenith correction to array with min "
        "solar zenith angle %s and max %s",
        sunzen_array.min(),
        sunzen_array.max(),
    )
    return numpy.ma.masked_array(
        (data_array / numpy.cos(numpy.deg2rad(sunzen_array))).data, origmask
    )


def apply_scale_factor(data_array, scale_factor):
    """Apply scale factor to all values in data_array.

    Scale factor applied as: data_array * scale_factor

    Parameters
    ----------
    data_array : numpy.ndarray or numpy.ma.MaskedArray
        data values to be scaled
    scale_factor : float
        requested scale factor

    Returns
    -------
    numpy.ndarray
        Return numpy.ndarray or numpy.ma.MaskedArray
        Input data array with scale factor applied
        data_array * scale_factor
    """
    LOG.info(
        "CORRECTION Applying scale factor of %s to array with min %s and max %s",
        scale_factor,
        data_array.min(),
        data_array.max(),
    )
    return data_array * scale_factor


def apply_offset(data_array, offset):
    """Apply offset to all values in data_array.

    Offset applied as: data_array + offset

    Parameters
    ----------
    data_array : numpy.ndarray or numpy.ma.MaskedArray
        data values to which offset will be applied.
    scale_factor : float
        requested offset.

    Returns
    -------
    numpy.ndarray
        Return numpy.ndarray or numpy.ma.MaskedArray
        Input data array with offset applied
        data_array + offset
    """
    LOG.info(
        "CORRECTION Applying offset of %s to array with min %s and max %s",
        offset,
        data_array.min(),
        data_array.max(),
    )
    return data_array + offset


def apply_minimum_value(data, min_val, outbounds):
    """Apply minimum values to an array of data.

    Parameters
    ----------
    data : numpy.ndarray or numpy.ma.MaskedArray
        data values to which the minimum value will be applied.
    min_val : float
        The minimum bound to be applied to the input data as a scalar.
    outbounds : str
        Method to use when applying bounds as a string. Valid values are:
          retain: keep all pixels as is
          mask: mask all pixels that are out of range.
          crop: set all out of range values to min_val.

    Returns
    -------
    numpy.ndarray
        Return numpy.ndarray or numpy.ma.MaskedArray
        Input data array with values below 'min_val'
        retained, cropped, or masked appropriately.
    """
    LOG.info("Applying minimum value of %r to data with min %f", min_val, data.min())

    # #Determine if mask is currently hardened
    # hardmask = data.hardmask
    # #Harden the mask to avoid unmasking bad values
    # if hardmask is False:
    #    data.harden_mask()

    # Allow for numpy arrays that are not masked arrays
    hardmask = None
    if hasattr(data, "hardmask"):
        # Determine if mask is currently hardened
        hardmask = data.hardmask
        # Harden the mask to avoid unmasking bad values
        if hardmask is False:
            data.harden_mask()

    # If outbounds is set to "mask" then mask the out of range data
    if outbounds == "mask":
        data = numpy.ma.masked_less(data, min_val)
    # If outbounds set to crop, then set out of range data to the minimum value
    elif outbounds == "crop":
        data[data < min_val] = min_val
    else:
        raise ValueError(
            'outbounds must be either "mask" or "crop".  Got %s.' % outbounds
        )

    # If the mask was originally not hardened, then unharden it now
    if hardmask is False:
        data.soften_mask()

    return data


def apply_maximum_value(data, max_val, outbounds):
    """Apply maximum value to an array of data.

    Parameters
    ----------
    data : numpy.ndarray or numpy.ma.MaskedArray
        data values to which the maximum value will be applied.
    max_val : float
        The maximum bound to be applied to the input data as a scalar.
    outbounds : str
        Method to use when applying bounds as a string. Valid values are:
          retain: keep all pixels as is
          mask: mask all pixels that are out of range.
          crop: set all out of range values to max_val.

    Returns
    -------
    numpy.ndarray
        Return numpy.ndarray or numpy.ma.MaskedArray
        Input data array with values above 'max_val'
        retained, cropped, or masked appropriately.
    """
    LOG.info("Applying maximum value of %r to data with max %f", max_val, data.max())

    # #Determine if mask is currently hardened
    # hardmask = data.hardmask
    # #Harden the mask to avoid unmasking bad values
    # if hardmask is False:
    #    data.harden_mask()

    # Allow for numpy arrays that are not masked arrays
    hardmask = None
    if hasattr(data, "hardmask"):
        # Determine if mask is currently hardened
        hardmask = data.hardmask
        # Harden the mask to avoid unmasking bad values
        if hardmask is False:
            data.harden_mask()

    # If outboudns is set to "mask" then mask the out of range data
    if outbounds == "mask":
        data = numpy.ma.masked_greater(data, max_val)
    # If outbounds is set to crop, then set out of range data to the maximum value
    elif outbounds == "crop":
        data[data > max_val] = max_val
    else:
        raise ValueError(
            'outbounds must be either "mask" or "crop".  Got %s.' % outbounds
        )

    # If the mask was originally not hardened, then unharden it now
    if hardmask is False:
        data.soften_mask()

    return data


def apply_data_range(
    data,
    min_val=None,
    max_val=None,
    min_outbounds="crop",
    max_outbounds="crop",
    norm=True,
    inverse=False,
):
    """
    Apply minimum and maximum values to an array of data.

    Normalize, invert, and handle out of bounds data as requested.

    Parameters
    ----------
    data : numpy.ndarray or numpy.ma.MaskedArray
        data values to which the data range will be applied.
    min_val : float, default None
        * The minimum bound to be applied to the input data as a scalar,
        * If None, use data.min().
    max_val : float, default=None
        * The maximum bound to be applied to the input data as a scalar.
        * If None, use data.max().
    min_outbounds : str, default='crop'
        Method to use when applying bounds as a string. Valid values are:

        * retain: keep all pixels as is
        * mask: mask all pixels that are out of range.
        * crop: set all out of range values to min_val
    max_outbounds : str, default='crop'
        Method to use when applying bounds as a string. Valid values are:

        * retain: keep all pixels as is
        * mask: mask all pixels that are out of range.
        * crop: set all out of range values to max_val
    norm : bool, default=True
        Boolean flag indicating whether to normalize (True) or not (False).

        * If True, returned data will be in the range from 0 to 1.
        * If False, returned data will be in the range min_val to max_val.
    inverse : bool, default=False
        Boolean flag indicating whether to invert data (True) or not (False).

        * If True, returned data will be inverted
        * If False, returned data will not be inverted

    Returns
    -------
    numpy.ndarray
        Return numpy.ndarray or numpy.ma.MaskedArray
        Input data array with values above 'max_val' or below 'min_val'
        retained, cropped, or masked.
    """
    # This is a temporary fix to treat xarrays like numpy arrays
    try:
        data = data.to_masked_array()
        is_xarr = True
    except AttributeError:
        pass
        is_xarr = False

    # Invert data if minimum value is greater than maximum value
    if inverse or (min_val is not None and max_val is not None and min_val > max_val):
        data, min_val, max_val = invert_data_range(data, min_val, max_val)

    # If a minimum value is specified, then apply minimum value
    if min_val is not None:
        data = apply_minimum_value(data, min_val, min_outbounds)
    else:
        min_val = data.min()

    # If a maximum value is specified, then apply maximum value
    if max_val is not None:
        data = apply_maximum_value(data, max_val, max_outbounds)
    else:
        max_val = data.max()

    # CHANGE DEFAULTS FOR OUTBOUNDS TO NONE??  If you want it to crop, specify.
    # Need to change in productfile/xml.py and utils/normalize.py
    # Actually, when plotting if you don't crop, it probably crops anyway.
    #   so maybe keep 'crop' as default anyway ?

    # Normalize data if requested
    if norm is True:
        data = normalize(data, min_val, max_val, min_outbounds, max_outbounds)

    if is_xarr:
        data = DataArray(data)

    return data


def invert_data_range(data, min_val=None, max_val=None):
    """
    Invert data range to an array of data.

    Parameters
    ----------
    data : numpy.ndarray or numpy.ma.MaskedArray
        data values to which the data range will be applied.
    min_val : float, optional
        The minimum bound to be applied to the input data as a scalar,
        by default None, which results in data.min().
    max_val : float, optional
        The maximum bound to be applied to the input data as a scalar.
        by default None, which results in data.max().

    Returns
    -------
    numpy.ndarray
        Return numpy.ndarray or numpy.ma.MaskedArray
        Input data array with values inverted.
    """
    if min_val is None:
        min_val = data.min()
    if max_val is None:
        max_val = data.max()
    LOG.info("Inverting data between %r and %r", min_val, max_val)

    if min_val > max_val:
        min_val, max_val = max_val, min_val

    if hasattr(data, "mask"):
        # Preserve mask
        origmask = data.mask
        data = numpy.ma.masked_array((max_val - (data - min_val)).data, origmask)
    else:
        data = max_val - (data - min_val)

    return data, min_val, max_val


def normalize(data, min_val=None, max_val=None, min_bounds="crop", max_bounds="crop"):
    """
    Normalize data array with min_val and max_val to range 0 to 1.

    Default to cropping outside requested data range.

    Parameters
    ----------
    data : numpy.ndarray or numpy.ma.MaskedArray
        data values to which the data range will be applied.
    min_val : float, default=None
        * The minimum bound to be applied to the input data as a scalar,
        * If None, use data.min().
    max_val : float, default=None
        * The maximum bound to be applied to the input data as a scalar.
        * If None, use data.max().
    min_outbounds : str, default='crop'
        Method to use when applying bounds as a string. Valid values are:

        * retain: keep all pixels as is
        * mask: mask all pixels that are out of range.
        * crop: set all out of range values to min_val
    max_outbounds : str, default='crop'
        Method to use when applying bounds as a string. Valid values are:

        * retain: keep all pixels as is
        * mask: mask all pixels that are out of range.
        * crop: set all out of range values to max_val

    Returns
    -------
    numpy.ndarray
        Return numpy.ndarray or numpy.ma.MaskedArray
        Input data array normalized between 0 and 1, with values above
        'max_val' or below 'min_val' retained, cropped, or masked.
    """
    # Determine if mask is currently hardened
    hardmask = None
    if hasattr(data, "hardmask"):
        hardmask = data.hardmask

    # Harden the mask to avoid unmasking bad values
    if hardmask is False:
        data.harden_mask()

    if min_val is None:
        min_val = data.min()
    if max_val is None:
        max_val = data.max()
    if min_bounds is None:
        min_bounds = "retain"
    if max_bounds is None:
        max_bounds = "retain"

    data = data.copy()
    data -= min_val
    data *= abs(1.0 / (max_val - min_val))

    if min_bounds == "retain":
        pass
    elif min_bounds == "crop":
        data[numpy.ma.where(data < 0.0)] = 0.0
    elif min_bounds == "mask":
        data = numpy.ma.masked_less(data, 0.0)

    if max_bounds == "retain":
        pass
    elif max_bounds == "crop":
        data[numpy.ma.where(data > 1.0)] = 1.0
    elif max_bounds == "mask":
        data = numpy.ma.masked_greater(data, 1.0)

    # If the mask was originally not hardened, then unharden it now
    if hardmask is False:
        data.soften_mask()

    return data


def apply_satellite_zenith_cutoff(data_array, satzen_array, satzen_angle_cutoff):
    """Apply satellite zenith angle cutoff to data_array.

    Mask values in data_array where satzen_array values exceed the cutoff threshold.

    Parameters
    ----------
    data_array : numpy.ndarray or numpy.ma.MaskedArray
        data values to be masked
    satzen_array : numpy.ndarray or numpy.ma.MaskedArray
        satellite zenith angles of the same shape as the data array.
    satzen_angle_cutoff : float
        Cutoff threshold for satellite zenith angle

    Returns
    -------
    numpy.ndarray
        Return numpy.ndarray or numpy.ma.MaskedArray if original
        data_array was MaskedArray with each value in the
        data_array divided by cos(sunzen).
    """
    LOG.info("Applying satellite zenith angle cutoff of %sdeg", satzen_angle_cutoff)
    # zen_mask = satzen_array > satzen_angle_cutoff
    return numpy.ma.masked_where(satzen_array > satzen_angle_cutoff, data_array)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Filename formatter for full-day text windspeed products."""
import logging

from os.path import join as pathjoin

from geoips.filenames.base_paths import PATHS as gpaths
from geoips.plugins.modules.filename_formatters.text_winds_full_fname import (
    assemble_windspeeds_text_full_fname,
)

LOG = logging.getLogger(__name__)

interface = "filename_formatters"
family = "xarray_metadata_to_filename"
name = "text_winds_day_fname"


def call(
    xarray_obj,
    extension=".txt",
    basedir=pathjoin(gpaths["ANNOTATED_IMAGERY_PATH"], "text_winds"),
):
    """Create full-day text windspeed filenames.

    text_winds_day_fname includes only YYYYMMDD in the filename, so all data for
    a full day is appended into a single file.

    See Also
    --------
    geoips.plugins.modules.filename_formatters.text_winds_full_fname.
        assemble_text_windspeeds_text_full_fname
        Shared utility for generating similarly formatted windspeed filenames.
    """
    return assemble_windspeeds_text_full_fname(
        basedir=basedir,
        source_name=xarray_obj.source_name,
        platform_name=xarray_obj.platform_name,
        data_provider=xarray_obj.data_provider,
        product_datetime=xarray_obj.start_datetime,
        dt_format="%Y%m%d",
        extension=extension,
        creation_time=None,
    )



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module to handle removing duplicate files, based on filename formats.

If an individual filename format has a method named
``"<filename_formatter>_remove_duplicates"``
defined, use that method to remove duplicates for the given current filename.
"""

import logging

LOG = logging.getLogger(__name__)


def remove_duplicates(fnames, remove_files=False):
    """Remove duplicate files from all filenames included in dict fnames.

    Parameters
    ----------
    fnames : dict
        Dictionary with individual filenames as keys, and a field named
        "filename_formatter" which indicates the filename format used to
        generate the given filename.
    remove_files : bool, optional
        Specify whether to remove files (True), or just list what would have
        been removed, default to False

    Returns
    -------
    removed_files : list
        List of files that were removed.
    saved_files : list
        List of files that were not removed.
    """
    removed_files = []
    saved_files = []
    from geoips.interfaces import filename_formatters
    from importlib import import_module

    for fname in fnames:
        if "filename_formatter" not in fnames[fname]:
            LOG.info("SKIPPING %s, no filename_formatter defined", fname)
            saved_files += [fname]
            continue
        filename_formatter = fnames[fname]["filename_formatter"]
        fname_fmt_plugin = filename_formatters.get_plugin(
            fnames[fname]["filename_formatter"]
        )
        if hasattr(
            import_module(fname_fmt_plugin.__module__),
            f"{filename_formatter}_remove_duplicates",
        ):
            fnamer_remove_dups = getattr(
                import_module(fname_fmt_plugin.__module__),
                f"{filename_formatter}_remove_duplicates",
            )
            curr_removed_files, curr_saved_files = fnamer_remove_dups(
                fname, remove_files=remove_files
            )
            removed_files += curr_removed_files
            saved_files += curr_saved_files
        else:
            LOG.warning(
                "SKIPPING DUPLICATE REMOVAL no "
                f"{filename_formatter}_remove_duplicates defined"
            )

    return removed_files, saved_files



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS interface module."""

# Note all imports from geoips.interfaces below will error on
# flake8 F401: "imported but unused", since we are not including the
# individual module name strings in the __all__ variable below.
# flake8 does not recognize the list of strings when passed to
# __all__.  Note F401 is ignored on this file in geoips/.config/flake8,
# so when using the "official" GeoIPS flake8 config, these errors will
# not be reported. Since flake8 does not allow you to specify a single
# error to ignore within the full file (only ALL errors within the file,
# or single errors on a single line), we are ignoring F401 in this file
# via a per-file ignore within the flake8 config.
# https://stackoverflow.com/questions/48153886/flake8-ignore-specific-warning-for-entire-file

from geoips.interfaces.module_based.algorithms import algorithms
from geoips.interfaces.module_based.colormappers import colormappers
from geoips.interfaces.module_based.output_checkers import output_checkers
from geoips.interfaces.module_based.coverage_checkers import coverage_checkers
from geoips.interfaces.module_based.databases import databases
from geoips.interfaces.module_based.filename_formatters import filename_formatters
from geoips.interfaces.module_based.interpolators import interpolators
from geoips.interfaces.module_based.output_formatters import (
    output_formatters,
)
from geoips.interfaces.module_based.procflows import procflows
from geoips.interfaces.module_based.readers import readers
from geoips.interfaces.module_based.sector_adjusters import (
    sector_adjusters,
)
from geoips.interfaces.module_based.sector_metadata_generators import (
    sector_metadata_generators,
)
from geoips.interfaces.module_based.sector_spec_generators import (
    sector_spec_generators,
)
from geoips.interfaces.module_based.title_formatters import (
    title_formatters,
)

from geoips.interfaces.yaml_based.feature_annotators import (
    feature_annotators,
)
from geoips.interfaces.yaml_based.gridline_annotators import (
    gridline_annotators,
)
from geoips.interfaces.yaml_based.product_defaults import product_defaults
from geoips.interfaces.yaml_based.products import products
from geoips.interfaces.yaml_based.sectors import sectors
from geoips.interfaces.yaml_based.workflows import workflows

# These lists are the "master" lists of the interface names.
# These are used in validating the plugins (ie, so we will catch a typo
# in an interface name)
module_based_interfaces = [
    "algorithms",
    "colormappers",
    "coverage_checkers",
    "databases",
    "filename_formatters",
    "interpolators",
    "output_checkers",
    "output_formatters",
    "procflows",
    "readers",
    "sector_adjusters",
    "sector_metadata_generators",
    "sector_spec_generators",
    "title_formatters",
]
yaml_based_interfaces = [
    "feature_annotators",
    "gridline_annotators",
    "product_defaults",
    "products",
    "sectors",
    "workflows",
]
# Note due to the fact that we are including all of the imported packages
# in __all__ via variables rather than the actual strings, flake8 does
# not recognize the above imports as being used.  F401 ignored via
# per-file ignore in geoips/.config/flake8 config.  See comment above
# for more information.
__all__ = module_based_interfaces + yaml_based_interfaces


def list_available_interfaces():
    """Return a dictionary of available interfaces.

    Collect and return a dictionary whose keys are the interface types (i.e.
    module_based, yaml_based, and text_based) and whose values are lists of the
    available interfaces for each interface type.
    """
    # These are buried to avoid polluting the interface module's namespace
    import inspect
    from geoips import interfaces

    all_interfaces = {
        "module_based": [],
        "text_based": [],
        "yaml_based": [],
    }
    for interface_type in ["module", "text", "yaml"]:
        try:
            available_interfaces = [
                str(mod_info[0])
                for mod_info in inspect.getmembers(
                    getattr(interfaces, f"{interface_type}_based"),
                    inspect.ismodule,
                )
            ]
            all_interfaces[f"{interface_type}_based"] = available_interfaces
        except AttributeError:
            continue

    return all_interfaces



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Passive Microwave 89 GHz Polarization Corrected Temperature.

Data manipulation steps for the "89pct" product.
This algorithm expects Brightness Temperatures in units of degrees Kelvin
"""
import logging

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "list_numpy_to_numpy"
name = "pmw_89pct"


def call(
    arrays,
    output_data_range,
    min_outbounds="crop",
    max_outbounds="mask",
    norm=False,
    inverse=False,
):
    """89pct product algorithm data manipulation steps.

    This algorithm expects Brightness Temperatures in units of degrees Kelvin,
    and returns degrees Kelvin

    Parameters
    ----------
    arrays : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data
            and other variables, in order of sensor "variables" list
        * Channel data: Degrees Kelvin

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of appropriately scaled channel data,
        in degrees Kelvin.
    """
    h89 = arrays[0]
    v89 = arrays[1]

    out = (1.7 * v89) - (0.7 * h89)

    from geoips.data_manipulations.corrections import apply_data_range

    data = apply_data_range(
        out,
        min_val=output_data_range[0],
        max_val=output_data_range[1],
        min_outbounds=min_outbounds,
        max_outbounds=max_outbounds,
        norm=norm,
        inverse=inverse,
    )

    return data



"""Lightweight helper for treating singular and plural spellings as equivalent.

A drop‑in str subclass that normalizes English nouns so that their
singular and plural forms compare as equal, hash to the same key, and work
inside Pydantic v2 models. Ignores capitalization.

Examples
--------
>>> from partial_lexeme import Lexeme
>>> Lexeme("reader") == "readers" == Lexeme("readers")
True
>>> {Lexeme("analyses"): 1} == {"analysis": 1}
True

>>> from pydantic import BaseModel
>>> class _Plugin(BaseModel):
>>>     kind: Lexeme
>>>     interface: Lexeme
>>> model = _Plugin(kind="reader", interface="readers")
>>> assert model.kind == model.interface == "reader"
"""

from __future__ import annotations
from typing import Any, Dict

__all__ = ["Lexeme"]

# Irregular nouns – plural ↔ singular
#   - Each plural maps to its singular
#   - Singulars map to themselves so that a single lookup works
_IRREGULAR: Dict[str, str] = {
    "analyses": "analysis",
    "analysis": "analysis",
    "indices": "index",
    "index": "index",
    "children": "child",
    "data": "datum",
    "criteria": "criterion",
}

# Reverse map – singular → plural
_S_TO_P: Dict[str, str] = {
    singular: plural for plural, singular in _IRREGULAR.items() if plural != singular
}


def _normalize(word: str) -> str:
    """Return the canonical singular, lower‑case form of word.

    Check the irregular table and then if not found
    apply heuristic suffix rules for regular nouns.
    """
    w = word.strip().lower()
    if not w:
        return ""

    if w in _IRREGULAR:
        return _IRREGULAR[w]

    # heuristics
    if w.endswith("ies") and len(w) > 3 and w[-4] not in "aeiou":
        return w[:-3] + "y"

    if w.endswith("es") and w[-3:] in {"ses", "xes", "zes"}:
        return w[:-2]

    if w.endswith("es") and w[-4:-2] in {"sh", "ch"}:
        return w[:-2]

    if w.endswith("s") and not w.endswith("ss") and len(w) > 3:
        return w[:-1]

    return w


def _to_plural(word: str) -> str:
    """Return a plausible plural spelling for *word*.

    This is best‑effort but the
    lexeme type does not depend on it because equality & hashing
    always normalizes back to singular.
    """
    w = word.strip().lower()

    if w in _S_TO_P:
        return _S_TO_P[w]

    # Simple heuristic rules
    if w.endswith("y") and w[-2] not in "aeiou":
        return w[:-1] + "ies"

    if w.endswith(("s", "x", "z")) or w.endswith("sh") or w.endswith("ch"):
        return w + "es"

    return w + "s"


class Lexeme(str):
    """A string that treats singular and plural spellings as equal.

    It behaves exactly like a built‑in ``str`` but overrides equality and
    hashing to use the canonical singular form.  This makes singular and
    plural spellings interchangeable as dict keys, set members, CLI options,
    etc.
    """

    def __new__(cls, value: Any) -> "Lexeme":
        """Create new Lexeme."""
        if isinstance(value, str):
            return super().__new__(cls, value)
        raise TypeError(
            f"{cls.__name__} must be built from str, got {type(value).__name__}"
        )

    @property
    def singular(self) -> "Lexeme":
        """Return the singular spelling."""
        return self.__class__(_normalize(self))

    @property
    def plural(self) -> "Lexeme":
        """Return a plural spelling."""
        return self.__class__(_to_plural(self.singular))

    def _key(self) -> str:
        """Normalize key for comparisons and hashing."""
        return _normalize(self)

    def __eq__(self, other: object) -> bool:
        """Plural or singular equality check."""
        if isinstance(other, str):
            return self._key() == _normalize(other)
        if isinstance(other, Lexeme):
            return self._key() == other._key()
        return False

    def __hash__(self) -> int:
        """Hash lexeme."""
        return hash(self._key())

    def __repr__(self) -> str:
        """Print out un-serialized self."""
        return f"{self.__class__.__name__}({super().__str__()!r})"

    @classmethod
    def __get_pydantic_core_schema__(cls, _source_type, _handler):
        """Return a core_schema that allows ``Lexeme`` for use as a field in pydantic.

        Accepts ``str`` or ``Lexeme``; coerces to ``Lexeme``.
        Always serializes output as plain string.
        """
        # lazy import so only imported if using pydantic
        from pydantic_core import core_schema

        def _to_lexeme(value: object, _info=None):
            if isinstance(value, cls):
                return value
            if isinstance(value, str):
                return cls(value)
            raise TypeError("String or Lexeme required")

        return core_schema.no_info_after_validator_function(
            _to_lexeme,
            core_schema.str_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda v: str(v)
            ),
        )



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Modules to access TC tracks, based on locations found in the deck files."""

import logging

from geoips.interfaces import (
    sector_metadata_generators,
    sector_spec_generators,
    sectors,
)

LOG = logging.getLogger(__name__)

# If we ever revert back to numbered storm from named storm, we may need to include this
# list in "get_final_storm_name" rather than just INVEST
# UNNAMED_STORM_NAMES = ['invest', 'one', 'two', 'three', 'four', 'five', 'six',
#                        'seven', 'eight', 'nine', 'ten', 'eleven', 'twelve',
#                        'thirteen', 'fourteen', 'fifteen', 'sixteen', 'seventeen',
#                        'eighteen', 'nineteen', 'twenty', 'twenty-one']


def create_tc_sector_info_dict(
    clat,
    clon,
    synoptic_time,
    storm_year,
    storm_basin,
    storm_num,
    aid_type=None,
    storm_name=None,
    final_storm_name=None,
    deck_line=None,
    source_sector_file=None,
    pressure=None,
    vmax=None,
):
    """Create storm info dictionary from items.

    Parameters
    ----------
    clat : float
        center latitude of storm
    clon : float
        center longitude of storm
    synoptic_time : datetime.datetime
        time of storm location
    storm_year : int
        4 digit year of storm
    storm_basin : str
        2 digit basin identifier
    storm_num : int
        2 digit storm number
    aid_type : str, default=None
        type of TC aid (BEST, MBAM, etc)
    storm_name : str, default=None
        Common name of storm
    final_storm_name : str, default=None
        Final name found throughout entire track file (ie, if reprocessing,
        will ensure early storm locations are identified with final storm name)
    deck_line : str, default=None
        source deck line for storm information
    pressure : float, default=None
        minimum pressure
    vmax : float, default=None
        maximum wind speed

    Returns
    -------
    fields : dict
        Dictionary of sector information, as passed into function.
    """
    fields = {}
    fields["clat"] = clat
    fields["clon"] = clon
    fields["synoptic_time"] = synoptic_time
    fields["storm_year"] = storm_year
    fields["storm_basin"] = storm_basin
    fields["storm_num"] = int(storm_num)
    fields["storm_name"] = "unknown"
    fields["aid_type"] = "unknown"
    fields["final_storm_name"] = "unknown"
    fields["source_sector_file"] = "unknown"
    if aid_type:
        fields["aid_type"] = aid_type
    if storm_name:
        fields["storm_name"] = storm_name
    if final_storm_name:
        LOG.info("USING passed final_storm_name %s", final_storm_name)
        fields["final_storm_name"] = final_storm_name
    else:
        LOG.info("USING storm_name as final_storm_name %s", storm_name)
        fields["final_storm_name"] = storm_name
    if source_sector_file:
        from geoips.geoips_utils import replace_geoips_paths

        fields["source_sector_file"] = replace_geoips_paths(source_sector_file)
    fields["pressure"] = pressure
    fields["deck_line"] = deck_line
    fields["vmax"] = vmax
    fields["source_filename"] = fields["source_sector_file"]
    fields["parser_name"] = None
    return fields


def get_tc_area_id(fields, finalstormname, tcyear):
    """Get TC area_id from fields, to be used as pyresample AreaDefinition area_id.

    Will be of form:
    * tcYYYYBBNNname (ie, tc2016io01one)
    """
    if not finalstormname:
        finalstormname = fields["storm_name"]
    newname = "{0}{1:02d}{2}".format(
        fields["storm_basin"].lower(), int(fields["storm_num"]), finalstormname.lower()
    )

    newname = newname.replace("_", "").replace(".", "").replace("-", "")

    # This ends up being tc2016io01one
    area_id = "tc" + str(tcyear) + newname
    return area_id


def get_tc_long_description(area_id, fields):
    """Return long_description of TC sector.

    This is commonly used as the long name/description on the
    pyresample AreaDefinition.
    """
    if "interpolated_time" in fields:
        long_description = "{0} interpolated_time {1}".format(
            area_id, str(fields["interpolated_time"])
        )
    else:
        long_description = "{0} synoptic_time {1}".format(
            area_id, str(fields["synoptic_time"])
        )
    return long_description


def set_tc_area_def(
    fields,
    tcyear=None,
    finalstormname=None,
    source_sector_file=None,
    clat=None,
    clon=None,
    tc_spec_template="tc_web",
    aid_type=None,
):
    """Set the TC area definition, using specified arguments.

    Parameters
    ----------
    fields : dict
        Dictionary of TC sector_info fields (clat, clon, storm name, etc)
        Valid fields can be found in geoips.sector_utils.utils.SECTOR_INFO_ATTRS
    tcyear : int, default=None
        Passed tcyear - since current year may not match tcyear for SHEM storms
    finalstormname : str, default=None
        finalstormname allows reprocessed storms to go in final storm directory
    source_sector_file : str, default=None
        attach source_sector_file to area_definition if known
    clat : float, default=None
        specify clat/clon separately from that found in 'fields'
    clon : float, default=None
        specify clat/clon separately from that found in 'fields'
    tc_spec_template: str, default="tc_web"
        Path to template YAML file to use when setting up area definition.
    aid_type : str, default=None
        type of TC aid (BEST, MBAM, etc)

    Returns
    -------
    pyresample.AreaDefinition
        pyresample AreaDefinition object with specified parameters.
    """
    if tc_spec_template is None:
        tc_spec_template = "tc_web"

    tc_template_plugin = sectors.get_plugin(tc_spec_template)

    # I think this is probably what we will want.
    template_func_name = tc_template_plugin["spec"]["sector_spec_generator"]["name"]
    template_args = tc_template_plugin["spec"]["sector_spec_generator"]["arguments"]

    if not finalstormname and "final_storm_name" in fields:
        finalstormname = fields["final_storm_name"]
    if not source_sector_file and "source_sector_file" in fields:
        source_sector_file = fields["source_sector_file"]
    if not source_sector_file and "source_filename" in fields:
        source_sector_file = fields["source_filename"]
    if not tcyear:
        tcyear = fields["storm_year"]
    if clat is None:
        clat = fields["clat"]
    if clon is None:
        clon = fields["clon"]

    area_id = get_tc_area_id(fields, finalstormname, tcyear)
    long_description = get_tc_long_description(area_id, fields)

    # These are things like 'center_coordinates'
    template_func = sector_spec_generators.get_plugin(template_func_name)
    # Probably generalize this at some point. For now I know those are the
    # ones that are <template>
    template_args["area_id"] = area_id
    template_args["long_description"] = long_description
    template_args["clat"] = clat
    template_args["clon"] = clon
    area_def = template_func(**template_args)

    if "interpolated_time" in fields:
        area_def.sector_start_datetime = fields["interpolated_time"]
        area_def.sector_end_datetime = fields["interpolated_time"]
    else:
        area_def.sector_start_datetime = fields["synoptic_time"]
        area_def.sector_end_datetime = fields["synoptic_time"]
    area_def.sector_type = "tc"
    area_def.sector_info = {}

    # area_def.description is Python3 compatible,
    # and area_def.name is Python2 compatible
    area_def.description = long_description
    if not hasattr(area_def, "name"):
        area_def.name = long_description

    from geoips.geoips_utils import replace_geoips_paths

    area_def.sector_info["source_sector_file"] = replace_geoips_paths(
        source_sector_file
    )
    # area_def.sector_info['sourcetemplate'] = dynamic_templatefname
    # area_def.sector_info['sourcedynamicxmlpath'] = dynamic_xmlpath
    # FNMOC sectorfile doesn't have pressure
    for fieldname in fields.keys():
        area_def.sector_info[fieldname] = fields[fieldname]
    area_def.sector_info["storm_year"] = tcyear

    # If storm_name is undefined in the current deck line, set it to finalstormname
    if area_def.sector_info["storm_name"] == "" and finalstormname:
        LOG.debug(
            'USING finalstormname "%s" rather than deck storm name "%s"',
            finalstormname,
            area_def.sector_info["storm_name"],
        )
        area_def.sector_info["storm_name"] = finalstormname
        area_def.sector_info["final_storm_name"] = finalstormname

    LOG.debug("      Current TC sector: %s", fields["deck_line"])
    return area_def


def trackfile_to_area_defs(
    trackfile_name, trackfile_parser="bdeck_parser", tc_spec_template=None
):
    """Get TC area definitions for the specified text trackfile.

    Limit to optionally specified trackfile_sectorlist

    Parameters
    ----------
    trackfile : str
        Full path to trackfile, convert each line into a separate area_def
    trackfile_parser : str
        Parser to use from plugins.modules.sector_metadata_generators on trackfiles

    Returns
    -------
    list
        List of pyresample AreaDefinition objects
    """
    if trackfile_parser is None:
        trackfile_parser = "bdeck_parser"

    parser = sector_metadata_generators.get_plugin(trackfile_parser)

    all_fields, final_storm_name, tc_year = parser(trackfile_name)

    area_defs = []
    LOG.info("STARTING setting TC area_defs")
    for fields in all_fields:
        # area_defs += [set_tc_sector(fields, dynamic_templatefname, finalstormname,
        #               tcyear, sfname, dynamic_xmlpath)]
        area_defs += [
            set_tc_area_def(
                fields,
                tc_year,
                finalstormname=final_storm_name,
                source_sector_file=trackfile_name,
                tc_spec_template=tc_spec_template,
            )
        ]
    LOG.info("FINISHED setting TC area_defs")

    return area_defs


def interpolate_storm_location(interp_dt, longitudes, latitudes, synoptic_times):
    """Interpolate the storm location at a specific time.

    Based on a list of known locations and times
    """
    LOG.info(
        "interp_dt: %s\nlatitudes:\n%s\nlongitudes:\n%s\nsynoptic_times:\n%s",
        interp_dt,
        latitudes,
        longitudes,
        synoptic_times,
    )



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Reader to read GFS Files in GRIB format.

    (NRL-Monterey, Sept. 2024)

    Current support includes:
       - GFS
       - GFS Wave global domain

This reader can take in multiple forcast times at multiple analysis times.
"""

# Python Standard Libraries
import logging
from os.path import basename


# Installed libraries
import xarray as xr
import pygrib as pyg
import numpy as np

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "gfs_grib"

surface_vars = [
    "Convective available potential energy",
    "Convective inhibition",
    "Temperature",
    "Snow depth",
    "Ice thickness",
    "Precipitation rate",
    "Vegetation",
    "Sunshine Duration",
    "Sea ice temperature",
    "Sea ice area fraction",
    "Wind speed (gust)",
]
# TBD: rename variables to shorter, simpler names
# ex: significant_height_of_combined_wind_waves_and_swell -> comb_wave_height
tmp_attrs = ["units", "pressureUnits"]
surf_attrs = ["units", "shortName", "name", "typeOfLevel"]

MS_TO_KTS = 1.94384


def convert_uv_to_kts(xr_df):
    """Convert U/V Components to windspeed and winddir."""
    sqrt_func = np.sqrt
    arctan2_func = np.arctan2
    degrees_func = np.degrees

    ucomp = xr_df.u_component_of_wind.values
    vcomp = xr_df.v_component_of_wind.values
    wind_dims = xr_df.u_component_of_wind.dims

    xr_df["windspeed_kts"] = xr.DataArray(
        sqrt_func((ucomp**2) + (vcomp**2)) * MS_TO_KTS, dims=wind_dims
    )
    xr_df["winddir"] = xr.DataArray(
        degrees_func(arctan2_func(-ucomp, -vcomp)) % 360, dims=wind_dims
    )
    return xr_df


def press_grb_to_xr(grb_obj, custom_pressure_name="unk_pres"):
    """Convert pressure grb values to xarray.Dataset.

    Parameters
    ----------
    grb_obj : list of pygrib objects
        * pygrib selected objects (pressure)

    Returns
    -------
    final_xarray : xarray.Dataset
        * xarray formatted to geoips standards
    """
    lats, lons = grb_obj[0].latlons()

    tmp_vars = set((i["name"] for i in grb_obj))
    varlist = {}
    unq_pres = False
    for i in tmp_vars:
        key = "_".join(i.lower().split(" "))
        tmp_data = next(j for j in grb_obj if j["name"] == i)

        data_attrs = {k: tmp_data[k] for k in tmp_attrs}
        tmp_press_lvls = sorted(set(j["level"] for j in grb_obj if j["name"] == i))

        if len(tmp_press_lvls) == 22:
            pres_dim = "coarse_press"
            cpres = tmp_press_lvls
        elif len(tmp_press_lvls) == 33:
            pres_dim = "fine_press"
            fpres = tmp_press_lvls
        else:
            pres_dim = custom_pressure_name
            upres = tmp_press_lvls
            unq_pres = True

        varlist[key] = xr.DataArray(
            data=[j.values for j in grb_obj if j["name"] == i],
            dims=(pres_dim, "xi", "yi"),
            attrs=data_attrs,
        )

    varlist["latitude"] = xr.DataArray(lats, dims=("xi", "yi"))
    varlist["longitude"] = xr.DataArray(lons, dims=("xi", "yi"))
    if unq_pres:
        varlist[custom_pressure_name] = xr.DataArray(upres, dims=(custom_pressure_name))
    else:
        varlist["coarse_press"] = xr.DataArray(cpres, dims=("coarse_press"))
        varlist["fine_press"] = xr.DataArray(fpres, dims=("fine_press"))

    xr_dset = xr.Dataset(varlist)

    return xr_dset


def surf_grb_to_xr(grb_obj):
    """Convert surface grb values to xarray.Dataset.

    Parameters
    ----------
    grb_obj : list of pygrib objects
        * pygrib selected objects (single/surface layer)

    Returns
    -------
    final_xarray : xarray.Dataset
        * xarray formatted to geoips standards
    """
    varlist = {}

    lats, lons = grb_obj[0].latlons()

    for i in grb_obj:
        tmp_key = "_".join(i.name.lower().split(" "))

        data_attrs = {k: i[k] for k in surf_attrs}

        varlist[tmp_key] = xr.DataArray(
            data=i.values, dims=("xi", "yi"), attrs=data_attrs
        )

    varlist["latitude"] = xr.DataArray(lats, dims=("xi", "yi"))
    varlist["longitude"] = xr.DataArray(lons, dims=("xi", "yi"))

    xr_dset = xr.Dataset(varlist)
    return xr_dset


def read_atmos(filenames):
    """Read gfs files.

    Parameters
    ----------
    filenames : list of filepaths
        * gfs grib files

    Returns
    -------
    final_xarray : xarray.Dataset
        * xarray formatted to geoips standards
    """
    LOG.info("Reading ATMOS file")
    xr_list = {"gfs_pressure": [], "gfs_surface": [], "gfs_singlelayer": []}
    base_time = []
    fcst_time = []
    for f in filenames:
        # read each file
        pg_frame = pyg.open(f)

        # Pressure
        tmp_pres = pg_frame.select(typeOfLevel="isobaricInhPa")
        xr_pres = press_grb_to_xr(tmp_pres)

        xr_pres = convert_uv_to_kts(xr_pres)

        # Surface
        tmp_surf = pg_frame.select(name=surface_vars, typeOfLevel="surface")
        surface_xr = surf_grb_to_xr(tmp_surf)

        # Single Layer, is ok to considerd as surface (slab)
        tmp_sl = pg_frame.select(typeOfLevel="atmosphereSingleLayer")
        single_atm_xr = surf_grb_to_xr(tmp_sl)

        xr_list["gfs_pressure"].append(xr_pres)
        xr_list["gfs_surface"].append(surface_xr)
        xr_list["gfs_singlelayer"].append(single_atm_xr)

        # track the analysis and valid forcast date
        base_time.append(tmp_sl[0].analDate)
        fcst_time.append(tmp_sl[0].validDate)

    tvar = xr.DataArray(base_time, dims=("atime"))
    bname = list(map(basename, filenames))
    res = float(".".join(bname[0].split(".")[3].split("p"))) * 111

    xr_attrs = {
        "source_file_name": bname,
        "start_datetime": base_time[0],
        "end_datetime": fcst_time[-1],
        "source_name": "gfs",
        "platform_name": "model",
        "data_provider": "NOAA",
        "sample_distance_km": res,
        "interpolation_radius_of_influence": 30000,
    }

    for k in xr_list.keys():
        xr_list[k] = xr.concat(xr_list[k], tvar)
        xr_list[k]["forcast_time"] = xr.DataArray(fcst_time, dims=("atime"))
        xr_list[k] = xr_list[k].assign_attrs(xr_attrs)
        # fix lat lon shape (fix upstream)
        xr_list[k]["latitude"] = xr_list[k]["latitude"][0]
        # if np.any(xr_list[k]['longitude'][0] > 180):
        xr_list[k]["longitude"] = xr_list[k]["longitude"][0]
        xr_list[k] = xr_list[k].transpose("xi", "yi", ...)

    return xr_list


def read_wave(filenames):
    """Read gfswave products."""
    LOG.info("Reading gfswave")
    xr_list = {"wave_surface": [], "wave_seq": []}
    base_time = []
    fcst_time = []
    for f in filenames:
        # read each file
        pg_frame = pyg.open(f)

        # Surface
        tmp_surf = pg_frame.select(typeOfLevel="surface")
        surface_xr = surf_grb_to_xr(tmp_surf)

        # sequence data
        tmp_seq = pg_frame.select(typeOfLevel="orderedSequenceData")
        # this is equal to levels
        seq_xr = press_grb_to_xr(tmp_seq, custom_pressure_name="wave_seq")

        xr_list["wave_seq"].append(seq_xr)
        xr_list["wave_surface"].append(surface_xr)

        base_time.append(tmp_seq[0].analDate)
        fcst_time.append(tmp_seq[0].validDate)

    tvar = xr.DataArray(base_time, dims=("atime"))
    bname = list(map(basename, filenames))
    res = float(".".join(bname[0].split(".")[3].split("p"))) * 111

    xr_attrs = {
        "source_file_name": bname,
        "start_datetime": base_time[0],
        "end_datetime": fcst_time[-1],
        "source_name": "gfs",
        "platform_name": "model",
        "data_provider": "NOAA",
        "sample_distance_km": res,
        "interpolation_radius_of_influence": 30000,
    }

    for k in xr_list.keys():
        xr_list[k] = xr.concat(xr_list[k], tvar)
        xr_list[k]["forcast_time"] = xr.DataArray(fcst_time, dims=("atime"))
        xr_list[k] = xr_list[k].assign_attrs(xr_attrs)
        # fix lat lon shape (fix upstream)
        xr_list[k]["latitude"] = xr_list[k]["latitude"][0]
        # if np.any(xr_list[k]['longitude'][0] > 180):
        xr_list[k]["longitude"] = xr_list[k]["longitude"][0]
        xr_list[k] = xr_list[k].transpose("xi", "yi", ...)

    return xr_list


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read GFS GRIB data.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.
        * Conforms to geoips xarray standards, see more in geoips documentation.
    """
    base_fnames = list(map(basename, fnames))
    fnames = sorted(fnames)

    tmp_read = pyg.open(fnames[0])
    tmp_val = tmp_read.read(1)[0]

    dtime_start = tmp_val.analDate
    dtime_end = tmp_val.validDate

    # rough resolution from degrees
    res = float(".".join(base_fnames[0].split(".")[3].split("p"))) * 111
    tmp_xr = xr.Dataset(
        attrs={
            "source_file_name": base_fnames,
            "start_datetime": dtime_start,
            "end_datetime": dtime_end,
            "source_name": "gfs",
            "platform_name": "model",
            "data_provider": "NOAA",
            "sample_distance_km": res,
            "interpolation_radius_of_influence": 30000,
        }
    )
    if metadata_only:
        tmp_dict = {"METADATA": tmp_xr}
        return tmp_dict

    gfs_type = set([i.split(".")[0] for i in base_fnames])
    xr_array = {}
    if len(gfs_type) > 1:
        raise FileExistsError(
            "Multiple file types given, please feed in one gfs file type."
        )
    if "gfs" in gfs_type:
        xr_array = read_atmos(fnames)
    elif "gfswave" in gfs_type:
        xr_array = read_wave(fnames)

    xr_array["METADATA"] = tmp_xr

    return xr_array



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Introspection functions on data arrays."""

import logging
import numpy

LOG = logging.getLogger(__name__)


def percent_unmasked(data_array):
    """Determine percent of a numpy.ma.Masked array that is not masked.

    Parameters
    ----------
    data_array : numpy.ma.MaskedArray
        Final processed array from which to determine coverage

    Returns
    -------
    float
        percent of input data array that is not masked.
    """
    return 100 * (float(numpy.ma.count(data_array)) / data_array.size)


def percent_not_nan(data_array):
    """Determine percent of a numpy.ndarray that is not NaN values.

    Parameters
    ----------
    data_array : numpy.ndarray
        Final processed array from which to determine coverage, invalid values
        specified by "numpy.nan".

    Returns
    -------
    float
        percent of input data array that is not numpy.nan.
    """
    return 100.0 * (
        1.0 - (numpy.count_nonzero(numpy.isnan(data_array)) / data_array.size)
    )



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test script for representative product comparisons."""

import logging

from geoips.commandline.log_setup import log_with_emphasis
from geoips.geoips_utils import get_numpy_seeded_random_generator

LOG = logging.getLogger(__name__)

interface = "output_checkers"
family = "standard"
name = "netcdf"


def get_test_files(test_data_dir):
    """Return a Series of Netcdf paths modified from compare."""
    import xarray as xr
    from os import makedirs
    from os.path import exists, join

    savedir = join(test_data_dir, "scratch", "unit_tests", "test_netcdf")
    if not exists(savedir):
        makedirs(savedir)
    # Path for the "compare" NetCDF file
    compare_file = join(savedir, "compare.nc")

    # Generate random data for the "compare" file
    predictable_random = get_numpy_seeded_random_generator()
    compare_data = predictable_random.random((100, 100))
    compare_ds = xr.Dataset(data_vars={"data": (("x", "y"), compare_data)})
    compare_ds.to_netcdf(compare_file)

    # Paths for the other files
    matched_path = join(savedir, "matched.nc")
    close_mismatch_path = join(savedir, "close_mismatch.nc")
    bad_mismatch_path = join(savedir, "bad_mismatch.nc")

    # Randomly modify the "compare" data for "close_mismatch" and "bad_mismatch"
    close_mismatch_data = compare_data + predictable_random.normal(
        scale=0.05, size=compare_data.shape
    )
    bad_mismatch_data = compare_data + predictable_random.normal(
        scale=0.25, size=compare_data.shape
    )

    # Create DataArrays for the modified data
    close_mismatch_da = xr.DataArray(data=close_mismatch_data, dims=("x", "y"))
    bad_mismatch_da = xr.DataArray(data=bad_mismatch_data, dims=("x", "y"))

    # Create Datasets and save them to NetCDF files
    close_mismatch_ds = xr.Dataset(data_vars={"data": close_mismatch_da})
    bad_mismatch_ds = xr.Dataset(data_vars={"data": bad_mismatch_da})

    close_mismatch_ds.to_netcdf(close_mismatch_path)
    bad_mismatch_ds.to_netcdf(bad_mismatch_path)
    compare_ds.to_netcdf(matched_path)
    return compare_file, [matched_path, close_mismatch_path, bad_mismatch_path]


def perform_test_comparisons(plugin, compare_file, test_files):
    """Test the comparison of two Netcdf files with the Netcdf Output Checker."""
    for path_idx in range(len(test_files)):
        retval = plugin.module.outputs_match(
            plugin,
            test_files[path_idx],
            compare_file,
        )
        if path_idx == 0:
            assert retval is True
        else:
            assert retval is False


def correct_file_format(fname):
    """Check if fname is a geoips formatted netcdf file.

    Parameters
    ----------
    fname : str
        Name of file to check.

    Returns
    -------
    bool
        True if it is a geoips netcdf file, False otherwise.
    """
    import xarray

    try:
        xobj = xarray.open_dataset(fname)
    except Exception:
        return False
    from geoips.geoips_utils import get_required_geoips_xarray_attrs

    return set(get_required_geoips_xarray_attrs()).issubset(set(xobj.attrs.keys()))


def outputs_match(plugin, output_product, compare_product):
    """Check if two geoips formatted netcdf files match.

    Parameters
    ----------
    plugin: OutputCheckerPlugin
        The corresponding geotiff OutputCheckerPlugin that has access to needed methods
    output_product : str
        Full path to current output product
    compare_product : str
        Full path to comparison product

    Returns
    -------
    bool
        Return True if products match, False if they differ
    """
    out_difftxt = plugin.get_out_diff_fname(compare_product, output_product)
    diffout = []
    retval = True
    import xarray

    out_xobj = xarray.open_dataset(output_product)
    compare_xobj = xarray.open_dataset(compare_product)

    if out_xobj.attrs != compare_xobj.attrs:
        message = "BAD GeoIPS NetCDF file attributes do NOT match exactly"
        log_with_emphasis(
            LOG.interactive,
            message,
            f"output_product: {output_product}",
            f"compare_product: {compare_product}",
        )
        for attr in out_xobj.attrs.keys():
            if attr not in compare_xobj.attrs:
                diffstr = (
                    f"\nattr {attr}\n\n"
                    f"output\n{out_xobj.attrs[attr]}\n\n"
                    "not in comparison\n"
                )
                diffout += [diffstr]
                LOG.interactive(diffstr)
            elif out_xobj.attrs[attr] != compare_xobj.attrs[attr]:
                diffstr = (
                    f"\nattr {attr}\n\n"
                    f"output\n{out_xobj.attrs[attr]}\n\n"
                    f"comparison\n{compare_xobj.attrs[attr]}\n"
                )
                diffout += [diffstr]
                LOG.interactive(diffstr)
        for attr in compare_xobj.attrs.keys():
            if attr not in out_xobj.attrs:
                diffstr = (
                    f"\nattr {attr}\n\n"
                    f"not in output\n\n"
                    f"comparison\n{compare_xobj.attrs[attr]}\n"
                )
                diffout += [diffstr]
                LOG.interactive(diffstr)
            elif out_xobj.attrs[attr] != compare_xobj.attrs[attr]:
                diffstr = (
                    f"\nattr {attr}\n\n"
                    f"output\n{out_xobj.attrs[attr]}\n\n"
                    f"comparison\n{compare_xobj.attrs[attr]}\n"
                )
                diffout += [diffstr]
                LOG.interactive(diffstr)
        diffout += ["\n"]
        retval = False

    if retval is False:
        with open(out_difftxt, "w") as fobj:
            fobj.writelines(diffout)
        return False

    try:
        xarray.testing.assert_allclose(compare_xobj, out_xobj)
    except AssertionError as resp:
        message = "BAD GeoIPS NetCDF files do not match within tolerance"
        log_with_emphasis(
            LOG.interactive,
            message,
            f"output_product: {output_product}",
            f"compare_product: {compare_product}",
        )
        log_with_emphasis(LOG.interactive, *[line for line in str(resp).split("\n")])
        diffout += [
            "\nxarray objects do not match between current output and comparison\n"
        ]
        diffout += [f"\nOut: {out_xobj}\n"]
        diffout += [f"\nCompare: {compare_xobj}\n"]
        diffout += [f"\n{resp}\n"]
        for varname in compare_xobj.variables:
            logged_messages = log_object_diff_values(
                out_xobj, compare_xobj, varname, log_function=LOG.warning
            )
            diffout.extend([[m] for m in logged_messages])
        retval = False

    try:
        xarray.testing.assert_identical(compare_xobj, out_xobj)
    except AssertionError as resp:
        lines = ["INFORMATIONAL ONLY assert_identical differences"]
        for line in str(resp).split("\n"):
            lines += [line]
        for varname in compare_xobj.variables:
            log_object_diff_values(
                out_xobj, compare_xobj, varname, log_function=LOG.info
            )
        log_with_emphasis(LOG.info, *lines)

    if retval is False:
        with open(out_difftxt, "w") as fobj:
            fobj.writelines(diffout)
        return False

    return True


def log_object_diff_values(object1, object2, compare_key, log_function=LOG.info):
    r"""
    Log differences two objects via key and returns messages detailing differences.

    Computes the maximum, minimum, and mean differences for the values associated
    with a specified key in two objects. Generates messages for any non-zero
    differences and logs these messages. The messages are also returned as a list
    of strings for further use. Takes in an optional custom logging function,
    defaults to info logging.

    Objects must expose a min, max, and mean function after the difference
    between them is computed.

    Parameters
    ----------
    object1 : object
        The first object containing the key for comparison. Must support
        subscripting with the compare key and arithmetic operations.
    object2 : object
        The second object for comparison. Must also support subscripting with the
        compare key and arithmetic operations.
    compare_key : str
        The key for the values to be compared between `object1` and `object2`.
    log_func : func, optional
        The function to be used for logging, must take in a list of strings of min
        length 0.

    Returns
    -------
    messages : list of str
        A list of messages indicating non-zero min, max, and mean differences for
        the values associated with `compare_key`, formatted as strings.

    Notes
    -----
    Only logs and returns messages for non-zero differences. If there are no non-zero
    differences, an empty list is returned.

    Uses `log_with_emphasis` to log the messages.

    Examples
    --------
    Given two objects `obj1` and `obj2` with a key `temp`:

    >>> obj1 = {'temp': np.array([1, 2, 3])}
    >>> obj2 = {'temp': np.array([2, 3, 4])}
    >>> log_object_diff_values(obj1, obj2, 'temp')
    ['mindiff temp: -1\\n', 'maxdiff temp: -1\\n', 'meandiff temp: -1.0\\n']

    This will log and return messages about differences in `temp` values.
    """
    maxdiff = (object2[compare_key] - object1[compare_key]).max()
    mindiff = (object2[compare_key] - object1[compare_key]).min()
    meandiff = (object2[compare_key] - object1[compare_key]).mean()
    messages = [
        f"mindiff {compare_key}: {mindiff}\n",
        f"maxdiff {compare_key}: {maxdiff}\n",
        f"meandiff {compare_key}: {meandiff}\n",
    ]
    log_with_emphasis(log_function, *messages)
    return messages


def call(plugin, compare_path, output_products):
    """Compare the "correct" netcdfs found the list of current output_products.

    Compares files produced in the current processing run with the list of
    "correct" files contained in "compare_path".

    Parameters
    ----------
    plugin: OutputCheckerPlugin
        The corresponding netcdf OutputCheckerPlugin that has access to needed methods
    compare_path : str
        Path to directory of "correct" products - filenames must match output_products
    output_products : list of str
        List of strings of current output products,
        to compare with products in compare_path

    Returns
    -------
    int
        Binary code: 0 if all comparisons were completed successfully.
    """
    retval = plugin.compare_outputs(
        compare_path,
        output_products,
    )
    return retval



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""PluginRegistry class to interface with the JSON plugin registries.

The "geoips config create-registries" utility generates a JSON file at the top
level of every geoips plugin package with a complete list of all plugins with
the associated metadata (everything except the actual contents of the plugin
itself).

Once all of the registered_plugins.json files have been generated via
geoips config create-registries, this class uses those registries to quickly
identify and open plugins as required.  Previously the individual
interface classes would open all plugins every time one was required,
so moving this process into a single PluginRegistry object allows us to
more effectively cache plugins across all interfaces, and avoid reading
in all plugins multiple times.
"""

from importlib import util, metadata, resources
from inspect import isclass
import logging
import os
from pathlib import Path
from types import SimpleNamespace

import json
import pydantic
import yaml

from geoips.create_plugin_registries import create_plugin_registries
from geoips.errors import PluginError, PluginRegistryError
from geoips.filenames.base_paths import PATHS
from geoips.geoips_utils import merge_nested_dicts


LOG = logging.getLogger(__name__)


class PluginRegistry:
    """Plugin Registry class definition.

    Represents all of the plugins found in all of the available GeoIPS packages.
    This class will load a plugin when requested, rather than loading all plugins when
    GeoIPS is instantiated.
    """

    def __init__(self, namespace, _test_registry_files=[]):
        """Initialize the plugin registry for the namespace provided.

        Where namespace is the group of plugin packages used to create the plugin
        registry.
        """
        self.namespace = namespace
        # Use this for unit testing
        if _test_registry_files:
            self.registry_files = _test_registry_files
            self._is_test = True
        # Use this for normal operation and collect the registry files
        else:
            self._is_test = False
            self.registry_files = self._find_registry_files(self.namespace)

    @property
    def registered_plugins(self):
        """Dictionary of every plugin's metadata found within self.namespace.

        self.namespace usually should correspond to 'geoips.plugin_packages' unless
        you're creating registries for plugins that exist outside this namespace.
        """
        if not hasattr(self, "_registered_plugins"):
            self._set_class_properties()
        return self._registered_plugins

    @registered_plugins.setter
    def registered_plugins(self, new_value):
        """Set the registered_plugins class attribute.

        See the registered_plugins property for more information.
        """
        self._registered_plugins = new_value

    @property
    def interface_mapping(self):
        """Dictionary of interface types and interfaces of that type.

        GeoIPS has three types of interfaces, though only two are commonly used
        (yaml_based, module_based). This dictionary has top level keys of all interface
        types, with their values being the list of unique interfaces that inherit that
        type.
        """
        if not hasattr(self, "_interface_mapping"):
            self._set_class_properties()
        return self._interface_mapping

    @interface_mapping.setter
    def interface_mapping(self, new_value):
        """Set the interface_mapping class attribute.

        See the interface_mapping property for more information.
        """
        self._registered_plugins = new_value

    @property
    def registered_yaml_based_plugins(self):
        """A dictionary of registered YAML-based plugins."""
        if not hasattr(self, "_registered_plugins"):
            self._set_class_properties()
        return self._registered_plugins["yaml_based"]

    @property
    def registered_module_based_plugins(self):
        """A dictionary of registered module-based plugins."""
        if not hasattr(self, "_registered_plugins"):
            self._set_class_properties()
        return self._registered_plugins["module_based"]

    def _set_class_properties(self, force_reset=False):
        """Find all plugins in registered plugin packages.

        Traverse the ``registered_plugins.json`` of each registered plugin package under
        self.namespace and add their entries to the _registered_plugins dictionary
        attribute.

        Parameters
        ----------
        force_reset: bool
            - Whether or not we want to force the plugin registry to recreate its
              'registered_plugins' attribute. This essentially forces a re-read of all
              of the registered_plugins.json files and recomputes the master dictionary.
            - Useful when we have rebuilt the registry files during runtime.
        """
        # Load the registries here and return them as a dictionary
        if not hasattr(self, "_registered_plugins") or force_reset:

            # Complete dictionary of all available plugins found in every geoips package
            self._registered_plugins = {}
            # A mapping of interfaces to plugin_types. Ie:
            # {
            # "yaml_based": [products, sectors, ...],
            # "module_based": [algorithms, readers, ...],
            # "text_based": [tpw_cimss, ...]
            # }
            self._interface_mapping = {}
            for reg_path in self.registry_files:
                try:
                    registry = self._load_registry(reg_path)
                except FileNotFoundError:
                    if PATHS["GEOIPS_REBUILD_REGISTRIES"]:
                        # This will be hit if we have this environment variable set to
                        # True
                        LOG.warning(
                            f"Plugin registry {reg_path} does not exist, "
                            "please run 'geoips config create-registries'"
                        )
                        # We attempt to create plugin registries under self.namespace
                        # if one or more plugin packages' registry file is missing and
                        # the GEOIPS_REBUILD_REGISTRIES environment is set to true. This
                        # should not be hit twice.

                        # Create plugin registries
                        self.create_registries()
                        registry = self._load_registry(reg_path)
                    else:
                        raise FileNotFoundError(
                            f"Plugin registry {reg_path} does not exist and "
                            "GEOIPS_REBUILD_REGISTRIES isn't set to True. To manually "
                            "create these files, run 'geoips config create-registries'."
                        )
                return_tuple = self._parse_registry(
                    self.interface_mapping, self.registered_plugins, registry
                )
                self.interface_mapping = return_tuple.interface_mapping
                self.registered_plugins = return_tuple.registered_plugins
            # Let's test this separately, not at runtime (see validate_all_registries).
            # Assume it was tested up front, and no longer needs testing at
            # runtime, so we don't fail catastrophically for a single bad
            # plugin that may not even be used.
            # if not self._is_test:
            #     self.validate_registry(
            #         self._registered_plugins,
            #         "all_registered_plugins",
            #     )

    @staticmethod
    def _find_registry_files(namespace):
        """Locate all plugin registry files found under 'namespace'.

        Parameters
        ----------
        namespace: str
            - The namespace in which plugin packages are registered to. Usually, this
              will be the 'geoips.plugin_package' namespace, however this can be changed
              by providing a different top-level namespace to the PluginRegistry class.

        Returns
        -------
        registry_files: list[str]
            - A list of filepaths corresponding to expected registered_plugins.json
              files for all plugin packages found under namespace.

        Raises
        ------
        PluginRegistryError:
            - Occurs if a package is potentially missing an __init__.py file.
        """
        registry_files = []  # Collect the paths to the registry files here
        for pkg in metadata.entry_points(group=namespace):
            try:
                registry_files.append(
                    str(resources.files(pkg.value) / "registered_plugins.json")
                )
            except TypeError:
                raise PluginRegistryError(
                    f"resources.files('{pkg.value}') failed\n"
                    f"pkg {pkg}\n"
                    "Potentially missing __init__.py file? Try:\n"
                    f"    touch {pkg.value}/__init__.py\n"
                    "and try again.\n"
                    "Note you will need to add a docstring to "
                    f"{pkg.value}/__init__.py in order for all tests to pass"
                )
        return registry_files

    @staticmethod
    def _load_registry(reg_path):
        """Load the plugin registry found at 'reg_path'.

        All files provided to this function share the same namespace. Usually, this
        will be the 'geoips.plugin_package' namespace, however this can be changed by
        providing a different top-level namespace to the PluginRegistry class.

        Parameters
        ----------
        reg_path: str
            - The absolute path to the plugin registry file for a certain plugin
              package.

        Returns
        -------
        registry: dict
            - A dictionary representing the contents of a plugin registry file. Can
              include top-level keys such as 'yaml_based' or 'module_based' each of
              which is another dictionary containing interfaces of that type, which
              are also dictionaries containing plugins that correspond to that
              interface.

        Raises
        ------
        FileNotFoundError:
            - Raised if 'reg_path' doesn't exist.
        """
        # if this is a yaml file, this is used for testing
        if Path(reg_path).suffix == ".yaml":
            with open(reg_path, "r") as fo:
                registry = yaml.safe_load(fo)
        else:
            with open(reg_path, "r") as fo:
                registry = json.load(fo)

        return registry

    @staticmethod
    def _parse_registry(interface_mapping, registered_plugins, registry):
        """Parse all plugins found under a package's plugin registry.

        Parameters
        ----------
        interface_mapping: dict
            - Dictionary of interface types and interfaces of that type.
              GeoIPS has three types of interfaces, though only two are commonly used
              (yaml_based, module_based). This dictionary has top level keys of all
              interface types, with their values being the list of unique interfaces
              that inherit that type.
        registered_plugins: dict
            - A dictionary of every plugin's metadata found within the plugin_registry's
              namespace.
        registry: dict
            - A dictionary representing the contents of a plugin registry file. Can
              include top-level keys such as 'yaml_based' or 'module_based' each of
              which is another dictionary containing interfaces of that type, which
              are also dictionaries containing plugins that correspond to that
              interface.

        Returns
        -------
        return_tuple: SimpleNamespace
            - A SimpleNamespace object containing updated values of the variables
              ['interface_mapping', 'registered_plugins'] which are dictionaries. See
              comments about those two variables in the parameters section above.
        """
        for plugin_type in registry:
            if plugin_type not in registered_plugins:
                registered_plugins[plugin_type] = {}
                interface_mapping[plugin_type] = []

            for interface in registry[plugin_type]:
                interface_dict = registry[plugin_type][interface]

                if interface not in registered_plugins[plugin_type]:
                    registered_plugins[plugin_type][interface] = interface_dict
                    interface_mapping[plugin_type].append(interface)
                else:
                    merge_nested_dicts(
                        registered_plugins[plugin_type][interface],
                        interface_dict,
                    )

        return_tuple = SimpleNamespace(
            interface_mapping=interface_mapping, registered_plugins=registered_plugins
        )
        return return_tuple

    def get_plugin_metadata(self, interface_obj, plugin_name):
        """Retrieve a plugin's metadata.

        Where the metadata of the plugin matches the plugin's corresponding entry in the
        plugin registry.

        Parameters
        ----------
        interface_obj: GeoIPS Interface Object
            - The object representing the interface class requesting plugin metadata.
        plugin_name: str or tuple(str)
            - The name of the plugin whose metadata we want.

        Returns
        -------
        metadata: dict
            - A dictionary of metadata for the requested plugin.
        """
        interface_registry = self.registered_plugins.get(
            interface_obj.interface_type, {}
        ).get(interface_obj.name)

        if interface_registry is None:
            raise KeyError(
                "Error: There is no interface in the plugin registry of type '"
                f"{interface_obj.interface_type}' called '{interface_obj.name}'."
            )

        if isinstance(plugin_name, tuple):
            # This occurs for product plugins: i.e. ('abi', 'Infrared')
            metadata = interface_registry.get(plugin_name[0], {}).get(plugin_name[1])
        elif isinstance(plugin_name, str):
            metadata = interface_registry.get(plugin_name)
        else:
            raise TypeError(
                f"Error: cannot search the plugin registry with the provided name = "
                f"{plugin_name}. Please provide either a string or a tuple of strings."
            )

        if metadata is None:
            raise PluginRegistryError(
                f"Error: There is no associated plugin under interface "
                f"'{interface_obj.name}' called '{plugin_name}'. If you're sure this "
                "plugin exists, please run 'geoips config create-registries'."
            )

        return metadata

    def get_yaml_plugin(self, interface_obj, name, rebuild_registries=None):
        """Get a YAML plugin by its name.

        Parameters
        ----------
        interface_obj: GeoIPS Interface Object
            - The object representing the interface class requesting this yaml plugin.
        name: str or tuple(str)
            - The name of the yaml-based plugin. Either a single string or a tuple of
              strings for product plugins.
        rebuild_registries: bool (default=None)
            - Whether or not to rebuild the registries if get_plugin fails. If set to
              None, default to what we have set in geoips.filenames.base_paths, which
              defaults to True. If specified, use the input value of rebuild_registries,
              which should be a boolean value. If rebuild registries is true and
              get_plugin fails, rebuild the plugin registry, call then call
              get_plugin once more with rebuild_registries toggled off, so it only gets
              rebuilt once.
        """
        try:
            registered_yaml_plugins = self.registered_plugins["yaml_based"]
        except KeyError:
            # Very likely could occur if registries haven't been built yet
            err_str = (
                "No YAML-based plugins found. There likely have been no plugin "
                "registries built yet. If automatic registry creation has been disabled"
                ", please run 'geoips config create-registries'."
            )
            self.retry_get_plugin(interface_obj, name, rebuild_registries, err_str)

        if rebuild_registries is None:
            rebuild_registries = interface_obj.rebuild_registries
        elif not isinstance(rebuild_registries, bool):
            raise TypeError(
                "Error: Argument 'rebuild_registries' was specified but isn't a boolean"
                f" value. Encountered this '{rebuild_registries}' instead."
            )

        interface_entry = registered_yaml_plugins[interface_obj.name]
        # This occurs for product plugins
        if isinstance(name, tuple):
            source_name = name[0]
            plg_name = name[1]
            extra_info = f"under source_name '{source_name}', "
        # This occurs for every other YAML plugin
        else:
            source_name = None
            plg_name = name
            extra_info = ""

        if plg_name not in interface_entry.get(source_name, interface_entry):
            err_str = (
                f"Plugin '{plg_name}', {extra_info}"
                f"from interface '{interface_obj.name}' "
                f"appears to not exist."
                f"\nCreate plugin, then call geoips config create-registries."
            )
            return self.retry_get_plugin(
                interface_obj, name, rebuild_registries, err_str
            )

        relpath = interface_entry.get(source_name, interface_entry)[plg_name]["relpath"]
        package = interface_entry.get(source_name, interface_entry)[plg_name]["package"]
        abspath = str(resources.files(package) / relpath)

        # If abspath doesn't exist the registry is out of date with the actual
        # contents of all, or a certain plugin package.
        if not os.path.exists(abspath):
            err_str = (
                f"Products plugin source: '{source_name}', plugin: '{plg_name} "
                f"exists in the registry but its corresponding file at '{abspath}' "
                "cannot be found. Reinstall your package and re-run "
                "'geoips config create-registries'."
            )
            # This error should never occur, but we're adding error handling here
            # just in case. The reason it will never occur is that, if the path
            # to such plugin does not exist, when geoips config create-registries is
            # re-run that syncs up the path to the associated plugin. It cannot
            # reach this point if the plugin name is invalid, so this point couldn't
            # be hit twice
            return self.retry_get_plugin(
                interface_obj,
                name,
                rebuild_registries,
                err_str,
                PluginRegistryError,
            )
        with open(abspath, "r") as file:
            documents = yaml.safe_load_all(file)
            plugin_found = False
            for plugin in documents:
                # This occurs for product plugins
                if source_name:
                    for product in plugin["spec"]["products"]:
                        if (
                            product["name"] == plg_name
                            and source_name in product["source_names"]
                        ):
                            plugin_found = True
                            plugin = product
                            break
                # Every other type of YAML plugin
                else:
                    if plugin["name"] == plg_name:
                        plugin_found = True

                if plugin_found:
                    break

        if not plugin_found:
            err_str = (
                f"Error: {interface_obj.name} YAML plugin under name '{name}' could"
                " not be found. Please ensure this plugin exists, and if it does, "
                "run 'geoips config create-registries'."
            )
            return self.retry_get_plugin(
                interface_obj, name, rebuild_registries, err_str
            )
        plugin["interface"] = interface_obj.name
        plugin["package"] = package
        plugin["abspath"] = abspath
        plugin["relpath"] = relpath

        if isclass(interface_obj.validator) and issubclass(
            pydantic.BaseModel, interface_obj.validator
        ):
            validated = interface_obj.validator(**plugin)
        else:
            validated = interface_obj.validator.validate(plugin)

        return interface_obj._plugin_yaml_to_obj(name, validated)

    def get_yaml_plugins(self, interface_obj):
        """Retrieve all yaml plugin objects for this interface.

        Parameters
        ----------
        interface_obj: GeoIPS Interface Object
            - The object representing the interface class requesting all plugins.
        """
        plugins = []
        registered_yaml_plugins = self.registered_yaml_based_plugins
        if interface_obj.name not in registered_yaml_plugins:
            LOG.debug("No plugins found for '%s' interface.", interface_obj.name)
            return plugins
        for name in registered_yaml_plugins[interface_obj.name].keys():
            plugins.append(self.get_yaml_plugin(interface_obj, name))
        return plugins

    def get_module_plugin(self, interface_obj, name, rebuild_registries=None):
        """Retrieve a plugin from this interface by name.

        Parameters
        ----------
        interface_obj: GeoIPS Interface Object
            - The object representing the interface class requesting this module plugin.
        name : str
            - The name the desired plugin.
        rebuild_registries: bool (default=None)
            - Whether or not to rebuild the registries if get_plugin fails. If set to
              None, default to what we have set in geoips.filenames.base_paths, which
              defaults to True. If specified, use the input value of rebuild_registries,
              which should be a boolean value. If rebuild registries is true and
              get_plugin fails, rebuild the plugin registry, call then call
              get_plugin once more with rebuild_registries toggled off, so it only gets
              rebuilt once.

        Returns
        -------
        An object of type ``<interface>Plugin`` where ``<interface>`` is the name of
        this interface.

        Raises
        ------
        PluginError
          If the specified plugin isn't found within the interface.
        """
        try:
            registered_module_plugins = self.registered_plugins["module_based"]
        except KeyError:
            # Very likely could occur if registries haven't been built yet
            err_str = f"No plugins found for '{interface_obj.name}' interface."
            self.retry_get_plugin(interface_obj, name, rebuild_registries, err_str)

        if rebuild_registries is None:
            rebuild_registries = interface_obj.rebuild_registries
        elif not isinstance(rebuild_registries, bool):
            raise TypeError(
                "Error: Argument 'rebuild_registries' was specified but isn't a boolean"
                f" value. Encountered this '{rebuild_registries}' instead."
            )

        if name not in registered_module_plugins[interface_obj.name]:
            err_str = (
                f"Plugin '{name}', "
                f"from interface '{interface_obj.name}' "
                f"appears to not exist."
                f"\nCreate plugin, then call geoips config create-registries."
            )
            return self.retry_get_plugin(
                interface_obj, name, rebuild_registries, err_str
            )

        package = registered_module_plugins[interface_obj.name][name]["package"]
        relpath = registered_module_plugins[interface_obj.name][name]["relpath"]
        module_path = os.path.splitext(relpath.replace("/", "."))[0]
        module_path = f"{package}.{module_path}"
        abspath = resources.files(package) / relpath
        # If abspath doesn't exist the registry is out of date with the actual contents
        # of all, or a certain plugin package.
        if not os.path.exists(abspath):
            err_str = (
                f"Plugin '{name}' exists in the registry but its corresponding file at "
                f"'{abspath}' cannot be found. Reinstall your package and re-run "
                "'geoips config create-registries'."
            )
            # This error should never occur, but we're adding error handling here
            # just in case. The reason it will never occur is that, if the path
            # to such plugin does not exist, when geoips config create-registries is
            # re-run that syncs up the path to the associated plugin. It cannot reach
            # this point if the plugin name is invalid, so this point couldn't be hit
            # twice
            return self.retry_get_plugin(
                interface_obj, name, rebuild_registries, err_str, PluginRegistryError
            )
        spec = util.spec_from_file_location(module_path, abspath)
        module = util.module_from_spec(spec)
        spec.loader.exec_module(module)
        plugin = interface_obj._plugin_module_to_obj(name, module)
        # This function might raise a PluginError with pertinent information on why
        # the plugin is invalid. Don't catch that, we want the error to be raised.
        interface_obj.plugin_is_valid(plugin)
        return plugin

    def get_module_plugins(self, interface_obj):
        """Retrieve all module plugins for this interface.

        Parameters
        ----------
        interface_obj: GeoIPS Interface Object
            - The object representing the interface class requesting all plugins.
        """
        plugins = []
        # All plugin interfaces are explicitly imported in
        # geoips/interfaces/__init__.py
        # self.name comes explicitly from one of the interfaces that are
        # found by default on geoips.interfaces.
        # If there is a defined interface with no plugins available in the current
        # geoips installation (in any currently installed plugin package),
        # then there will NOT be an entry within registered plugins
        # for that interface, and a KeyError will be raised in the for loop
        # below.
        # Check if the current interface (self.name) is found in the
        # registered_plugins dictionary - if it is not, that means there
        # are no plugins for that interface, so return an empty list.
        registered_module_plugins = self.registered_module_based_plugins
        if interface_obj.name not in registered_module_plugins:
            LOG.debug("No plugins found for '%s' interface.", interface_obj.name)
            return plugins

        for plugin_name in registered_module_plugins[interface_obj.name]:
            try:
                plugins.append(self.get_module_plugin(interface_obj, plugin_name))
            except AttributeError as resp:
                raise PluginError(
                    f"Plugin '{plugin_name}' is missing the 'name' attribute, "
                    f"\nfrom package '{plugin_name['package']},' "
                    f"'{plugin_name['relpath']}' module,"
                ) from resp
        return plugins

    def retry_get_plugin(
        self, interface_obj, name, rebuild_registries, err_str, err_type=PluginError
    ):
        """Rerun self.get_plugin, but call 'geoips config create-registries' beforehand.

        By running 'geoips config create-registries', we automate the registration of
        plugins in GeoIPS. If the plugin persists not to be found, then we'll raise an
        appropriate PluginError as denoted by 'err_str'.

        Parameters
        ----------
        interface_obj: GeoIPS Interface Object
            - The object representing the interface class requesting this plugin.
        name: str or tuple(str)
            - The name of the yaml plugin. Either a single string or a tuple of strings
              for product plugins.
        rebuild_registries: bool
            - Whether or not to rebuild the registries if get_plugin fails. If set to
              true and get_plugin fails, rebuild the plugin registry, call then call
              get_plugin once more with rebuild_registries toggled off, so it only gets
              rebuilt once.
        err_str: string
            - The error to be reported.
        err_type: Exception-based Class
            - The class of exception to be raised.
        """
        if rebuild_registries:
            LOG.interactive(
                "Running 'geoips config create-registries' due to a missing plugin "
                f"located under interface: '{interface_obj.name}', plugin_name: "
                f"'{name}'."
            )
            self.create_registries()
            # Force a rebuild of the master 'registered_plugins' dictionary.
            self._set_class_properties(force_reset=True)
            # This is done as some interface classes override 'get_plugin' with
            # additional parameters it it's call signature. We only want to call
            # BaseYamlInterface or BaseModuleInterface 'get_plugin', then return such
            # information to the child class which overrode 'get_plugin'. Implementing
            # it this way ensures that will happen.
            base_interface_class = interface_obj.__class__.__base__()
            base_interface_class.name = interface_obj.name
            return base_interface_class.get_plugin(name, rebuild_registries=False)
        else:
            raise err_type(err_str)

    def create_registries(self, packages=None, save_type="json") -> None:
        """Create one or more plugin registry files.

        By default, this command will create all plugin registry files for all
        installed geoips packages (geoips.plugin_packages entrypoint). If packages is
        provided via the argument above, create registry files associated with each
        of those packages.

        Parameters
        ----------
        packages: list[str], default=None
            - A list of names corresponding to geoips.plugin_packges whose registries
              we want to create.
        save_type: str, default="json"
            - Format to write registries to. This will also be the file extension. Valid
              options are either 'json' or 'yaml'.

        Raises
        ------
        ValueError:
            - Raised if 'save_type' is provided but is not one of ['json', 'yaml']
        TypeError:
            - Raised if packages is provided and is not a list of strings.
        PluginRegistryError:
            - Raised if one or more of the packages provided is not a valid
              geoips.plugin_package, or if the associated namespace provided is not
              a valid namespace.
        """
        if save_type not in ["json", "yaml"]:
            raise ValueError(
                "Error: 'save_type' kwarg was provided but is not one of "
                "['json', 'yaml']."
            )
        plugin_packages = metadata.entry_points(group=self.namespace)
        if len(plugin_packages) == 0:
            raise PluginRegistryError(
                f"Error: namespace '{self.namespace}' does not exist or has no plugin "
                "packages associated with it. Please resolve this before continuing."
            )
        if packages:
            self._validate_packages_input(packages)
            filtered_packages = []
            for pkg in plugin_packages:
                if pkg.value in packages:
                    filtered_packages.append(pkg)
            plugin_packages = filtered_packages

        LOG.debug(plugin_packages)
        create_plugin_registries(plugin_packages, save_type, self.namespace)

    def delete_registries(self, packages=None) -> None:
        """Delete one or more plugin registry files.

        By default, this command will delete all plugin registry files found in all
        installed geoips packages (geoips.plugin_packages entrypoint). If packages is
        provided via the argument above, delete the registry file(s) associated with
        each of those packages.

        Parameters
        ----------
        packages: list[str], default=None
            - A list of names corresponding to geoips.plugin_packges whose registries
              we want to delete.

        Raises
        ------
        TypeError:
            - Raised if packages is provided and is not a list of strings.
        FileNotFoundError:
            - Raised if a registry file could not be found in one or more
              geoips.plugin_packages.
        PluginRegistryError:
            - Raised if one or more of the packages provided is not a valid
              geoips.plugin_package, or if the associated namespace provided is not
              a valid namespace.
        """
        if packages:
            self._validate_packages_input(packages)
        else:
            packages = [ep.value for ep in metadata.entry_points(group=self.namespace)]
            if len(packages) == 0:
                raise PluginRegistryError(
                    f"Error: namespace '{self.namespace}' does not exist or has no "
                    "plugin packages associated with it. Please resolve this before "
                    "continuing."
                )

        for pkg in packages:
            # If packages is provided and the current package is in that list, or if
            # we're using the default value for that argument, delete the associated
            # registry
            if (packages and pkg in packages) or packages is None:
                yaml_plug_path = str(resources.files(pkg) / "registered_plugins.yaml")
                json_plug_path = str(resources.files(pkg) / "registered_plugins.json")
                for path in [json_plug_path, yaml_plug_path]:
                    # Attempt to remove the files, pass silently if they don't exist.
                    try:
                        os.remove(path)
                    except FileNotFoundError:
                        continue

    def _validate_packages_input(self, packages):
        """Validate that packages is a list of strings.

        If not, then raise a TypeError indicating what was formatted incorrectly.

        Parameters
        ----------
        packages: list[str], default=None
            - A list of names corresponding to geoips.plugin_packges whose registries
              we want to delete.

        Raises
        ------
        TypeError:
            - Raised if packages is provided and is not a list of strings.
        PluginRegistryError:
            - Raised if one or more of the packages provided is not a valid
              geoips.plugin_package, or if the associated namespace provided is not
              a valid namespace.
        """
        if not isinstance(packages, list):
            raise TypeError(
                "Error: 'packages' kwarg was provided but it is not a list object."
            )
        elif any([not isinstance(pkg, str) for pkg in packages]):
            raise TypeError(
                "Error: 'packages' kwarg was provided but one or more of it's "
                "items were not a string."
            )
        # If any package name in 'packages' could not be associated with an entry point
        # found in pypi's package registry associated with 'self.namespace', raise a
        # plugin registry error.
        found_packages = [
            pkg.value for pkg in metadata.entry_points(group=self.namespace)
        ]
        if any([pkg_name not in found_packages for pkg_name in packages]):
            raise PluginRegistryError(
                "Error: either the namespace provided was invalid or one or more of "
                "the packages whose registry you requested to delete does not exist."
            )


plugin_registry = PluginRegistry("geoips.plugin_packages")



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Standard xarray-based NetCDF output format."""

import logging

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "xarray_data"
name = "netcdf_xarray"


def call(xarray_obj, product_names, output_fnames, clobber=False):
    """Write xarray-based NetCDF outputs to disk."""
    for ncdf_fname in output_fnames:
        write_xarray_netcdf(xarray_obj, ncdf_fname)
    return output_fnames


def write_xarray_netcdf(xarray_obj, ncdf_fname, clobber=False):
    """Write out xarray_obj to netcdf file named ncdf_fname."""

    def check_attr(xobj, attr):
        """Check xarray attributes."""
        if isinstance(xobj.attrs[attr], datetime):
            xobj.attrs[attr] = xobj.attrs[attr].strftime("%c")
        if xobj.attrs[attr] is None:
            xobj.attrs[attr] = str(xobj.attrs[attr])
        if isinstance(xobj.attrs[attr], bool):
            xobj.attrs[attr] = str(xobj.attrs[attr])
        # DO NOT attempt to store dicts - they are not supported
        if isinstance(xobj.attrs[attr], dict):
            LOG.warning(
                "SKIPPING attr %s %s, dictionaries not supported for netcdf output",
                attr,
                xobj.attrs[attr],
            )
            xobj.attrs.pop(attr)

    from geoips.filenames.base_paths import make_dirs
    from os.path import dirname

    make_dirs(dirname(ncdf_fname))

    orig_attrs = xarray_obj.attrs.copy()
    orig_var_attrs = {}

    from datetime import datetime

    for attr in xarray_obj.attrs.copy().keys():
        check_attr(xarray_obj, attr)
    for varname in xarray_obj.variables.keys():
        orig_var_attrs[varname] = xarray_obj[varname].attrs.copy()
        for attr in xarray_obj[varname].attrs.keys():
            check_attr(xarray_obj[varname], attr)

    roi_str = "none"
    if "interpolation_radius_of_influence" in xarray_obj.attrs.keys():
        roi_str = xarray_obj.interpolation_radius_of_influence

    sdt_str = "none"
    if "start_datetime" in xarray_obj.attrs.keys():
        sdt_str = xarray_obj.attrs["start_datetime"]

    edt_str = "none"
    if "end_datetime" in xarray_obj.attrs.keys():
        edt_str = xarray_obj.attrs["end_datetime"]

    dp_str = "none"
    if "data_provider" in xarray_obj.attrs.keys():
        dp_str = xarray_obj.attrs["data_provider"]

    area_def_str = "none"
    # GEOIPS 1 COMPATIBILITY
    if "area_def" in xarray_obj.attrs.keys():
        # Must pop off the actual area_defintion - does not write to xarray properly
        area_def = xarray_obj.attrs.pop("area_def")
        area_def_str = repr(area_def)
        xarray_obj.attrs["area_def_str"] = area_def_str
    # If actual area_definition object, write it out to xarray as str
    elif "area_definition" in xarray_obj.attrs.keys():
        # If area_definition_str was explicitly defined on the area_definition
        # object, use that
        if hasattr(xarray_obj.area_definition, "area_definition_str"):
            # Must pop off the actual area_defintion - does not write to xarray properly
            area_def_str = xarray_obj.area_definition.area_definition_str
            area_def = xarray_obj.attrs.pop("area_definition")
            xarray_obj.attrs["area_definition_str"] = area_def_str
        else:
            # Must pop off the actual area_defintion - does not write to xarray properly
            area_def = xarray_obj.attrs.pop("area_definition")
            area_def_str = repr(area_def)
            xarray_obj.attrs["area_definition_str"] = area_def_str

    LOG.info(
        "Writing xarray obj to file %s, source %s, platform %s, start_dt %s, end_dt %s,"
        " %s %s, %s %s, %s %s",
        ncdf_fname,
        xarray_obj.source_name,
        xarray_obj.platform_name,
        sdt_str,
        edt_str,
        "provider",
        dp_str,
        "roi",
        roi_str,
        "area_def",
        area_def_str,
    )

    from os.path import exists

    if clobber is True or not exists(ncdf_fname):
        xarray_obj.to_netcdf(ncdf_fname)
    else:
        LOG.warning("SKIPPING not outputing file %s, exists", ncdf_fname)
    xarray_obj.attrs = orig_attrs
    for varname in xarray_obj.variables.keys():
        xarray_obj[varname].attrs = orig_var_attrs[varname]

    return [ncdf_fname]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Derived Motion Winds at vertical pressure levels plotted as Barbs in Knots or m s-1.

Data manipulation steps for derived motion winds products.
This algorithm expects derived motion wind speeds at various pressures
"""
import logging

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "xarray_to_xarray"
name = "windbarbs_dmw"


def call(
    xobj,
    variables,
    product_name,
    output_data_range=None,
    pressure_level_range=None,
    input_units=None,
    output_units=None,
    min_outbounds="crop",
    max_outbounds="crop",
    norm=False,
    inverse=False,
    var_map=None,
):
    """Windbarbs product algorithm data manipulation steps.

    This algorithm expects input windspeed with units in either "m s-1" or "kts",
    and returns in either "m s-1" or "kts" (as specified in product YAML).

    Parameters
    ----------
    xobj : xarray.Dataset
        * Dataset holding at minimum wind speed, direction, and pressure data
    variables : list of str
        * list of input variables used in algorithm
    product_name : str
        * Name used to store output of algorithm to xobj
    output_data_range : list of float, default=None
        * list of min and max value for wind speeds (kts or m s-1)
        * defaults to None, which results in using data.min and data.max.
    pressure_level_range : list of float, default=None
        * list of min and max pressure levels to filter derived motion wind retrievals.
        * defaults to None, which results in using all wind retrievals
    input_units : str, default=None
        * Units of input data, for applying necessary conversions
        * defaults to None, resulting in no unit conversions.
    output_units : str, default=None
        * Units of output data, for applying necessary conversions
        * defaults to None, resulting in no unit conversions.
    min_outbounds : str, default='crop'
        * Method to use when applying bounds.  Valid values are:

            * retain: keep all pixels as is
            * mask: mask all pixels that are out of range
            * crop: set all out of range values to either min_val or max_val
              as appropriate
    max_outbounds : str, default='crop'
        * Method to use when applying bounds.  Valid values are:

            * retain: keep all pixels as is
            * mask: mask all pixels that are out of range
            * crop: set all out of range values to either min_val or max_val
              as appropriate
    norm : bool, default=False
        * Boolean flag indicating whether to normalize (True) or not (False)

            * If True, returned data will be in the range from 0 to 1
            * If False, returned data will be in the range from min_val to
              max_val
    inverse : bool, default=False
        * Boolean flag indicating whether to inverse (True) or not (False)

            * If True, returned data will be inverted
            * If False, returned data will not be inverted
    var_map : dict, default=None
        * Dictionary that maps input variables to names used in xobj

    Returns
    -------
    xarray.Dataset with new data variable named after product_name
        numpy.ndarray or numpy.MaskedArray of appropriately scaled channel data,
        dstacked as follows:

        * (spd, direction, rain_flag)
        * spd in kts
        * direction in degrees
        * rain_flag 0 (for compatibility with windbarb output formatter)
    """
    import numpy

    if not var_map:
        var_map = {}

    spd = xobj[var_map.get("speed", "wind_speed")]
    direction = xobj[var_map.get("direction", "wind_direction")]
    pressure = xobj[var_map.get("pressure", "pressure")]
    if output_data_range is None:
        output_data_range = (float(spd.min()), float(spd.max()))
    if pressure_level_range is None:
        pressure_level_range = (float(pressure.min()), float(pressure.max()))

    from geoips.data_manipulations.conversions import unit_conversion

    spd = unit_conversion(spd, input_units, output_units)

    from geoips.data_manipulations.corrections import apply_data_range

    spd = apply_data_range(
        spd,
        min_val=output_data_range[0],
        max_val=output_data_range[1],
        min_outbounds=min_outbounds,
        max_outbounds=max_outbounds,
        norm=norm,
        inverse=inverse,
    )
    pressure_mask = (pressure.data >= pressure_level_range[0]) & (
        pressure.data <= pressure_level_range[1]
    )

    stacked = numpy.dstack(
        (
            spd.where(pressure_mask),
            direction.where(pressure_mask),
            numpy.zeros(spd.shape),
            pressure.where(pressure_mask),
        )
    ).squeeze()
    stacked_dims = ("dim_0", "dim_1", "dim_2")[: stacked.ndim - 1]
    xobj[product_name] = ((*stacked_dims, product_name), stacked)
    xobj[product_name].attrs["windbarb_data_columns"] = [
        "speed",
        "direction",
        "rain_flag",
        "pressure",
    ]
    return xobj



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""MIMIC TPW NetCDF reader."""

# Python Standard Libraries
from datetime import datetime
import logging
import os

# Third-Party Libraries
import numpy
import xarray

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "mimic_netcdf"
source_names = ["mimic"]


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read TPW MIMIC data from a list of filenames.

    Dataset information::

        <xarray.Dataset>
        Dimensions:             (lat: 721, lon: 1440)
        Dimensions without coordinates: lat, lon
        Data variables:
            lonArr              (lon) float32 ...
            latArr              (lat) float32 ...
            tpwGrid             (lat, lon) float32 ...
            tpwGridPrior        (lat, lon) float32 ...
            tpwGridSubseq       (lat, lon) float32 ...
            timeAwayGridPrior   (lat, lon) timedelta64[ns] ...
            timeAwayGridSubseq  (lat, lon) timedelta64[ns] ...
            footGridPrior       (lat, lon) float32 ...
            footGridSubseq      (lat, lon) float32 ...
            satGridPrior        (lat, lon) uint8 ...
            satGridSubseq       (lat, lon) uint8 ...

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    fname = fnames[0]

    xobj = xarray.open_dataset(fname)

    [date, time, ext] = os.path.basename(fname).split(".")
    dt = datetime.strptime(date + time, "comp%Y%m%d%H%M%S")
    xobj.attrs["data_provider"] = "cimss"

    xobj.attrs["start_datetime"] = dt
    xobj.attrs["end_datetime"] = dt
    xobj.attrs["platform_name"] = "tpw"
    xobj.attrs["source_file_datetimes"] = [dt]
    xobj.attrs["source_name"] = "mimic"
    # ~2km for data_fine
    xobj.attrs["sample_distance_km"] = 110.0 * abs(
        xobj.variables["latArr"][1].data - xobj.variables["latArr"][0].data
    )
    xobj.attrs["interpolation_radius_of_influence"] = (
        xobj.attrs["sample_distance_km"] * 1000.0 * 1.5
    )
    if metadata_only is True:
        return {"METADATA": xobj}

    LOG.info("Obtaining lat/lon from xarray")
    # Meshgrid requires Latitude and Longitude at once, so don't put in loop
    lat = xobj.variables["latArr"][...]
    lon = xobj.variables["lonArr"][...]

    LOG.info("Calculating full lat/lon grid")
    lon_final, lat_final = numpy.meshgrid(lon, lat)

    LOG.info("Adding lat grid to xarray")
    xobj["latitude"] = xarray.DataArray(numpy.ma.array(lat_final), dims=("lat", "lon"))
    LOG.info("Adding lon grid to xarray")
    xobj["longitude"] = xarray.DataArray(numpy.ma.array(lon_final), dims=("lat", "lon"))
    xobj = xobj.drop("latArr")
    xobj = xobj.drop("lonArr")
    xobj["tpw"] = xobj["tpwGrid"]

    return {"MIMIC": xobj, "METADATA": xobj[[]]}



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `test script` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

import pytest

from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


class TestGeoipsTestScript(BaseCliTest):
    """Unit Testing Class for Test Script Command."""

    @property
    def command_combinations(self):
        """A subset list of all possible commands used by the GeoipsTestScript command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            self._cmd_list = []
            base_args = ["geoips", "test", "script"]
            # Add argument list executing a dummy test script
            self._cmd_list.append(base_args + ["cli_dummy_script.sh"])
            # Do the same thing specifying which package it comes from
            self._cmd_list.append(base_args + ["-p", "geoips", "cli_dummy_script.sh"])
            # Add argumetn list executing a dummy integration test
            self._cmd_list.append(
                base_args
                + [
                    "--integration",
                    "cli_dummy_integration.sh",
                ],
            )
            # Do the same thing specifying which package it comes from
            self._cmd_list.append(
                base_args
                + [
                    "-p",
                    "geoips",
                    "--integration",
                    "cli_dummy_integration.sh",
                ],
            )
            # Add argument list to retrieve help message
            self._cmd_list.append(base_args + ["-h"])
            # Add argument list with non existent package
            self._cmd_list.append(
                base_args
                + [
                    "-p",
                    "non_existent_package",
                    "abi.static.Infrared.imagery_annotated.sh",
                ]
            )
            # Add argument list with non existent script name in default geoips pkg
            self._cmd_list.append(base_args + ["non_existent_script_name"])
            # Add arg list to run integration tests from another package (will fail)
            self._cmd_list.append(
                base_args + ["-p", "data_fusion", "--integration", "base_test.sh"]
            )
        return self._cmd_list

    def check_error(self, args, error):
        """Ensure that the 'geoips test script ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        editable = self.assert_non_editable_error_or_wrong_package(args, error)
        if editable:
            # An error occurred using args. Assert that args is not valid and check the
            # output of the error.
            assert (
                "geoips test script -p <package_name> <--integration> <script_name>`"
                in error
            )

    def check_output(self, args, output):
        """Ensure that the 'geoips test script ...' successful output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        # The args provided are valid, so test that the output is actually correct
        if "-h" in args:
            assert (
                "geoips test script -p <package_name> <--integration> <script_name>`"
                in output
            )
        elif "--integration" in args:
            assert "I ran a dummy integration test." in output
        else:
            # Checking that output from geoips test script command reports succeeds
            assert "I ran a dummy test script." in output


test_sub_cmd = TestGeoipsTestScript()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips test script ...' commands.

    This test covers every valid combination of commands for the 'geoips test script'
    command. We also test invalid commands, to ensure that the proper help documentation
    is provided for those using the command incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with (ie. ['geoips', 'test', 'script'])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Coverage check routine for center radius coverage checks."""

import logging

import numpy
from skimage.draw import disk

LOG = logging.getLogger(__name__)

interface = "coverage_checkers"
family = "standard"
name = "center_radius"


def plot_coverage(main_ax, area_def, covg_args):
    """Plot the coverage specified by the 'center_radius' function.

    Parameters
    ----------
    main_ax : matplotlib.axis
        Axis on which to plot coverage representation
    area_def : pyresample.AreaDefinition
        area def for current plot
    covg_args : dict
        product params dictionary for current product -
        to ensure we plot the correct coverage params

    Returns
    -------
    No return value
    """
    plot_color = "black"

    if "radius_km" in covg_args:
        radius_km = covg_args["radius_km"]
        res_km = max(area_def.pixel_size_x, area_def.pixel_size_y) / 1000.0
        radius_pixels = 1.0 * radius_km / res_km
        main_ax.scatter(
            0, 0, s=2 * radius_pixels**2, facecolors="none", edgecolors=plot_color
        )


def create_radius(temp_arr, radius_pixels=300, x_center=0, y_center=0):
    """Create a radius around given x,y coordinates in the 2d array.

    Given the radius and the x,y coordinates it creates a circle around those
    points using the skimage.draw library

    Parameters
    ----------
    temp_arr : int
        The 2D array.
    radius : int, optional
        The radius of the circle. 500 is default value.
    x : int, optional
        The x coordinate of middle circle point. 0 is default value.
    y : int, optional
        The x coordinate of middle circle point. 0 is default value.

    Returns
    -------
    numpy.ndarray
        2D array with circle created at the x,y coordinate with the given radius
        All circles are marked as 1.
    """
    dumby_arr = numpy.zeros((temp_arr.shape), dtype=numpy.uint8)
    r_points, c_points = disk(
        (x_center, y_center), radius_pixels, shape=dumby_arr.shape
    )

    dumby_arr[r_points, c_points] = 1

    return dumby_arr


def call(
    xarray_obj,
    variable_name,
    area_def=None,
    radius_km=300,
):
    """Coverage check routine for xarray objects with masked projected arrays.

    Parameters
    ----------
    xarray_obj : xarray.Dataset
        xarray object containing variable "variable_name"
    variable_name : str
        variable name to check percent unmasked
    radius_km : float
        Radius of center disk to check for coverage

    Returns
    -------
    float
        Percent coverage of variable_name
    """
    if variable_name not in xarray_obj:
        raise KeyError(
            f"Variable {variable_name} did not exist. Can not calculate coverage."
        )

    temp_arr = xarray_obj[variable_name].to_masked_array()

    res_km = (
        min(
            xarray_obj.area_definition.pixel_size_x,
            xarray_obj.area_definition.pixel_size_y,
        )
        / 1000.0
    )
    radius_pixels = 1.0 * radius_km / res_km
    LOG.info(
        "Using %s km radius, %s pixels radius, "
        "%s km resolution, variable %s, area_def %s",
        radius_km,
        radius_pixels,
        res_km,
        variable_name,
        area_def,
    )

    dumby_arr = create_radius(
        temp_arr,
        radius_pixels=radius_pixels,
        x_center=temp_arr.shape[0] / 2,
        y_center=temp_arr.shape[1] / 2,
    )

    num_valid_in_radius = numpy.count_nonzero(
        numpy.where(dumby_arr & ~temp_arr.mask, 1, 0)
    )
    num_total_in_radius = numpy.count_nonzero(dumby_arr)

    return (float(num_valid_in_radius) / num_total_in_radius) * 100.0



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Routines for outputting formatted text wind speed and vector data files."""

import logging
import os
from datetime import datetime
import numpy
import shutil

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "xrdict_varlist_outfnames_to_outlist"
name = "text_winds"


def call(
    xarray_dict, varlist, output_fnames, append=False, overwrite=True, source_names=None
):
    """Write text windspeed output file."""
    output_products = []
    num_arrs = 0
    for key in xarray_dict:
        if key == "METADATA":
            continue
        xarray_obj = xarray_dict[key]
        if "wind_speed_kts" not in xarray_obj.variables:
            continue
        if num_arrs == 0:
            curr_append = append
        else:
            curr_append = True
        num_arrs = num_arrs + 1
        output_products += write_text_winds(
            xarray_obj,
            varlist,
            output_fnames.copy(),  # It will overwrite it if we don't copy
            append=curr_append,
            overwrite=overwrite,
            source_names=source_names,
        )

    # Remove any duplicates - they would have been overwritten
    return list(set(output_products))


def write_text_winds(
    xarray_obj, varlist, output_fnames, append=False, overwrite=True, source_names=None
):
    """Write out TC formatted text file of wind speeds.

    Parameters
    ----------
    text_fname : str
        String full path to output filename
    speed_array : ndarray
        array of windspeeds
    time_array : ndarray
        array of POSIX time stamps same length as speed_array
    lon_array : ndarray
        array of longitudes of same length as speed_array
    lat_array : ndarray
        array of latitudes of same length as speed_array
    platform_name : str
        String platform name
    """
    # NOTE long does not exist in Python 3, so changed this to int.  This will
    # limit us to 32 bit integers within Python 2
    # time_array = wind_xarray['time'].to_masked_array().astype(long).flatten()
    time_array = xarray_obj["time"].to_masked_array().astype(int).flatten()
    # This results in an array of POSIX timestamps - seconds since epoch.
    time_array = time_array / 10**9

    speed_array = xarray_obj["wind_speed_kts"].to_masked_array().flatten()
    lat_array = xarray_obj["latitude"].to_masked_array().flatten()
    lon_array = xarray_obj["longitude"].to_masked_array().flatten()
    platform_name = xarray_obj.platform_name
    source_name = platform_name.upper()
    # If we passed a dictionary of source_name mappings, and it contains the
    # current platform_name, swap it out.
    if source_names is not None and platform_name in source_names:
        source_name = source_names[platform_name].upper()
    dir_array = None
    if "wind_dir_deg_met" in xarray_obj.variables:
        dir_array = xarray_obj["wind_dir_deg_met"].to_masked_array().flatten()
    output_products = []

    if not isinstance(platform_name, str):
        raise TypeError("Parameter platform_name must be a str")
    if not isinstance(output_fnames, list):
        raise TypeError("Parameter output_fnames must be a list of str")
    if not isinstance(speed_array, numpy.ndarray):
        raise TypeError("Parameter speed_array must be a numpy.ndarray of wind speeds")
    if not isinstance(lat_array, numpy.ndarray):
        raise TypeError("Parameter lat_array must be a numpy.ndarray of latitudes")
    if not isinstance(lon_array, numpy.ndarray):
        raise TypeError("Parameter lon_array must be a numpy.ndarray of longitudes")
    if not isinstance(time_array, numpy.ndarray):
        raise TypeError(
            "Parameter time_array must be a numpy.ndarray of POSIX timestamps"
        )

    if hasattr(speed_array, "mask"):
        if dir_array is not None:
            newmask = (
                speed_array.mask
                | time_array.mask
                | lat_array.mask
                | lon_array.mask
                | dir_array.mask
            )
        else:
            newmask = (
                speed_array.mask | time_array.mask | lat_array.mask | lon_array.mask
            )
        inds = numpy.ma.where(~newmask)
        speed_array = speed_array[inds]
        time_array = time_array[inds]
        lon_array = lon_array[inds]
        lat_array = lat_array[inds]
        if dir_array is not None:
            dir_array = dir_array[inds]

    openstr = "w"
    if append:
        openstr = "a"
    startdt_str = datetime.utcfromtimestamp(time_array[0]).strftime("%Y%m%d%H%M")
    header = ""

    for text_fname in output_fnames:
        if os.path.exists(text_fname) and not overwrite and not append:
            LOG.info("File already exists, not overwriting %s", text_fname)
            continue
        if not isinstance(text_fname, str):
            raise TypeError("Parameter text_fname must be a str")

        from geoips.filenames.base_paths import make_dirs

        make_dirs(os.path.dirname(text_fname))

    text_fname = output_fnames.pop()
    if not os.path.exists(text_fname) or not append:
        header = "METXSCT {0} ASC (FULL DAY)\n".format(startdt_str)
    with open(text_fname, openstr) as fobj:
        if dir_array is not None:
            fobj.write(header)
            for speed, time, lon, lat, direction in zip(
                speed_array, time_array, lon_array, lat_array, dir_array
            ):
                # dtstr = time.strftime('%Y%m%d%H%M')
                dtstr = datetime.utcfromtimestamp(time).strftime("%Y%m%d%H%M")
                # if lon > 180:
                #     lon = lon - 360
                format_string = " {0:>3s} {1:>8.1f} {2:>6.1f} {3:>3d} {4:>3d} {5:s}\n"
                fobj.write(
                    format_string.format(
                        source_name, lat, lon, int(direction), int(speed), dtstr
                    )
                )
        else:
            for speed, time, lon, lat in zip(
                speed_array, time_array, lon_array, lat_array
            ):
                # dtstr = time.strftime('%Y%m%d%H%M')
                dtstr = datetime.utcfromtimestamp(time).strftime("%Y%m%d%H%M")
                # if lon > 180:
                #     lon = lon - 360
                format_string = "{0:>6s}{1:>8.2f}{2:>8.2f}{3:>4d} {4:s}\n"
                if source_name == "SMAP" or source_name == "SMOS":
                    format_string = " {0:<6s} {1:>5.1f} {2:>5.1f} {3:>3d} {4:s}\n"
                fobj.write(
                    format_string.format(source_name, lat, lon, int(speed), dtstr)
                )
        output_products += [text_fname]

    ctime = datetime.fromtimestamp(os.stat(text_fname).st_ctime)
    LOG.info("    SUCCESS wrote out text windspeed file %s at %s", text_fname, ctime)

    for additional_text_fname in output_fnames:
        shutil.copy(text_fname, additional_text_fname)
        ctime = datetime.fromtimestamp(os.stat(additional_text_fname).st_ctime)
        LOG.info(
            "    SUCCESS wrote out text windspeed file %s at %s",
            additional_text_fname,
            ctime,
        )
        output_products += [additional_text_fname]
    return output_products
